from __future__ import print_function, division, absolute_import
import os
import unittest
import pathlib

import numpy as np
from numpy.testing import assert_almost_equal

try:
    import matplotlib
except ImportError:
    matplotlib = None

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse
from openmdao.utils.general_utils import set_pyoptsparse_opt

import dymos as dm
from dymos.examples.hyper_sensitive.hyper_sensitive_ode import HyperSensitiveODE
from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE
from dymos.examples.brachistochrone.brachistochrone_vector_states_ode import BrachistochroneVectorStatesODE
from dymos.utils.misc import om_version
from dymos.utils.testing_utils import _get_reports_dir

_, optimizer = set_pyoptsparse_opt('IPOPT', fallback=True)


@use_tempdirs
class TestRunProblem(unittest.TestCase):

    @require_pyoptsparse(optimizer='IPOPT')
    def test_run_HS_problem_radau_hp_refine(self):
        p = om.Problem(model=om.Group())
        p.driver = om.pyOptSparseDriver()
        p.driver.declare_coloring()
        p.driver.options['optimizer'] = optimizer

        if optimizer == 'SNOPT':
            p.driver.opt_settings['Major iterations limit'] = 200
            p.driver.opt_settings['Major feasibility tolerance'] = 1.0E-6
            p.driver.opt_settings['Major optimality tolerance'] = 1.0E-6
        elif optimizer == 'IPOPT':
            p.driver.opt_settings['hessian_approximation'] = 'limited-memory'
            p.driver.opt_settings['print_level'] = 0
            p.driver.opt_settings['max_iter'] = 200
            p.driver.opt_settings['linear_solver'] = 'mumps'

        traj = p.model.add_subsystem('traj', dm.Trajectory())
        phase0 = traj.add_phase('phase0', dm.Phase(ode_class=HyperSensitiveODE,
                                                   transcription=dm.Radau(num_segments=10,
                                                                          order=3)))
        phase0.set_time_options(fix_initial=True, fix_duration=True)
        phase0.add_state('x', fix_initial=True, fix_final=False, rate_source='x_dot', targets=['x'])
        phase0.add_state('xL', fix_initial=True, fix_final=False, rate_source='L', targets=['xL'])
        phase0.add_control('u', opt=True, targets=['u'], rate_continuity=False)

        phase0.add_boundary_constraint('x', loc='final', equals=1)

        phase0.add_objective('xL', loc='final')

        phase0.set_refine_options(refine=True, tol=1e-6)

        p.setup(check=True)

        tf = 20.0

        p.set_val('traj.phase0.states:x', phase0.interp('x', [1.5, 1]))
        p.set_val('traj.phase0.states:xL', phase0.interp('xL', [0, 1]))
        p.set_val('traj.phase0.t_initial', 0)
        p.set_val('traj.phase0.t_duration', tf)
        p.set_val('traj.phase0.controls:u', phase0.interp('u', [-0.6, 2.4]))
        dm.run_problem(p, refine_method='hp', refine_iteration_limit=10)

        sqrt_two = np.sqrt(2)
        val = sqrt_two * tf
        c1 = (1.5 * np.exp(-val) - 1) / (np.exp(-val) - np.exp(val))
        c2 = (1 - 1.5 * np.exp(val)) / (np.exp(-val) - np.exp(val))

        ui = c1 * (1 + sqrt_two) + c2 * (1 - sqrt_two)
        uf = c1 * (1 + sqrt_two) * np.exp(val) + c2 * (1 - sqrt_two) * np.exp(-val)
        J = 0.5 * (c1 ** 2 * (1 + sqrt_two) * np.exp(2 * val) + c2 ** 2 * (1 - sqrt_two) *
                   np.exp(-2 * val) - (1 + sqrt_two) * c1 ** 2 - (1 - sqrt_two) * c2 ** 2)

        assert_near_equal(p.get_val('traj.phase0.timeseries.u')[0],
                          ui,
                          tolerance=5e-4)

        assert_near_equal(p.get_val('traj.phase0.timeseries.u')[-1],
                          uf,
                          tolerance=5e-4)

        assert_near_equal(p.get_val('traj.phase0.timeseries.xL')[-1],
                          J,
                          tolerance=5e-4)

    @require_pyoptsparse(optimizer='IPOPT')
    def test_run_HS_problem_radau_ph_refine(self):
        p = om.Problem(model=om.Group())
        p.driver = om.pyOptSparseDriver()
        p.driver.declare_coloring()
        p.driver.options['optimizer'] = optimizer

        if optimizer == 'SNOPT':
            p.driver.opt_settings['Major iterations limit'] = 200
            p.driver.opt_settings['Major feasibility tolerance'] = 1.0E-6
            p.driver.opt_settings['Major optimality tolerance'] = 1.0E-6
        elif optimizer == 'IPOPT':
            p.driver.opt_settings['print_level'] = 0
            p.driver.opt_settings['max_iter'] = 200
            p.driver.opt_settings['linear_solver'] = 'mumps'
            p.driver.opt_settings['mu_strategy'] = 'monotone'
            p.driver.opt_settings['bound_mult_init_method'] = 'mu-based'
            p.driver.opt_settings['mu_init'] = 0.01

        traj = p.model.add_subsystem('traj', dm.Trajectory())
        phase0 = traj.add_phase('phase0', dm.Phase(ode_class=HyperSensitiveODE,
                                                   transcription=dm.Radau(num_segments=10,
                                                                          order=3)))
        phase0.set_time_options(fix_initial=True, fix_duration=True)
        phase0.add_state('x', fix_initial=True, fix_final=False, rate_source='x_dot', targets=['x'])
        phase0.add_state('xL', fix_initial=True, fix_final=False, rate_source='L', targets=['xL'])
        phase0.add_control('u', opt=True, targets=['u'], rate_continuity=False)

        phase0.add_boundary_constraint('x', loc='final', equals=1)

        phase0.add_objective('xL', loc='final')

        phase0.set_refine_options(refine=True, tol=1e-6)

        p.setup(check=True)

        tf = 20.0

        p.set_val('traj.phase0.states:x', phase0.interp('x', [1.5, 1]))
        p.set_val('traj.phase0.states:xL', phase0.interp('xL', [0, 1]))
        p.set_val('traj.phase0.t_initial', 0)
        p.set_val('traj.phase0.t_duration', tf)
        p.set_val('traj.phase0.controls:u', phase0.interp('u', [-0.6, 2.4]))
        dm.run_problem(p, refine_method='ph', refine_iteration_limit=10)

        sqrt_two = np.sqrt(2)
        val = sqrt_two * tf
        c1 = (1.5 * np.exp(-val) - 1) / (np.exp(-val) - np.exp(val))
        c2 = (1 - 1.5 * np.exp(val)) / (np.exp(-val) - np.exp(val))

        ui = c1 * (1 + sqrt_two) + c2 * (1 - sqrt_two)
        uf = c1 * (1 + sqrt_two) * np.exp(val) + c2 * (1 - sqrt_two) * np.exp(-val)
        J = 0.5 * (c1 ** 2 * (1 + sqrt_two) * np.exp(2 * val) + c2 ** 2 * (1 - sqrt_two) *
                   np.exp(-2 * val) - (1 + sqrt_two) * c1 ** 2 - (1 - sqrt_two) * c2 ** 2)

        assert_near_equal(p.get_val('traj.phase0.timeseries.u')[0],
                          ui,
                          tolerance=5e-4)

        assert_near_equal(p.get_val('traj.phase0.timeseries.u')[-1],
                          uf,
                          tolerance=5e-4)

        assert_near_equal(p.get_val('traj.phase0.timeseries.xL')[-1],
                          J,
                          tolerance=5e-4)

    @require_pyoptsparse(optimizer='IPOPT')
    def test_run_HS_problem_gl(self):
        p = om.Problem(model=om.Group())
        p.driver = om.pyOptSparseDriver()
        p.driver.declare_coloring()
        p.driver.options['optimizer'] = optimizer

        if optimizer == 'SNOPT':
            p.driver.opt_settings['Major iterations limit'] = 100
            p.driver.opt_settings['Major feasibility tolerance'] = 1.0E-6
            p.driver.opt_settings['Major optimality tolerance'] = 1.0E-6
        elif optimizer == 'IPOPT':
            p.driver.opt_settings['hessian_approximation'] = 'limited-memory'
            # p.driver.opt_settings['nlp_scaling_method'] = 'user-scaling'
            p.driver.opt_settings['print_level'] = 0
            p.driver.opt_settings['linear_solver'] = 'mumps'

        traj = p.model.add_subsystem('traj', dm.Trajectory())
        phase0 = traj.add_phase('phase0', dm.Phase(ode_class=HyperSensitiveODE,
                                                   transcription=dm.GaussLobatto(num_segments=20,
                                                                                 order=3)))
        phase0.set_time_options(fix_initial=True, fix_duration=True)
        phase0.add_state('x', fix_initial=True, fix_final=False, rate_source='x_dot', targets=['x'])
        phase0.add_state('xL', fix_initial=True, fix_final=False, rate_source='L', targets=['xL'])
        phase0.add_control('u', opt=True, targets=['u'], rate_continuity=False)

        phase0.add_boundary_constraint('x', loc='final', equals=1)

        phase0.add_objective('xL', loc='final')

        phase0.set_refine_options(refine=True, tol=1.0E-5)

        p.setup(check=True)

        tf = 20

        p.set_val('traj.phase0.states:x', phase0.interp('x', [1.5, 1]))
        p.set_val('traj.phase0.states:xL', phase0.interp('xL', [0, 1]))
        p.set_val('traj.phase0.t_initial', 0)
        p.set_val('traj.phase0.t_duration', tf)
        p.set_val('traj.phase0.controls:u', phase0.interp('u', [-0.6, 2.4]))
        dm.run_problem(p, refine_method='hp', refine_iteration_limit=5)

        sqrt_two = np.sqrt(2)
        val = sqrt_two * tf
        c1 = (1.5 * np.exp(-val) - 1) / (np.exp(-val) - np.exp(val))
        c2 = (1 - 1.5 * np.exp(val)) / (np.exp(-val) - np.exp(val))

        ui = c1 * (1 + sqrt_two) + c2 * (1 - sqrt_two)
        uf = c1 * (1 + sqrt_two) * np.exp(val) + c2 * (1 - sqrt_two) * np.exp(-val)
        J = 0.5 * (c1 ** 2 * (1 + sqrt_two) * np.exp(2 * val) + c2 ** 2 * (1 - sqrt_two) *
                   np.exp(-2 * val) - (1 + sqrt_two) * c1 ** 2 - (1 - sqrt_two) * c2 ** 2)

        assert_near_equal(p.get_val('traj.phase0.timeseries.u')[0],
                          ui,
                          tolerance=1e-2)

        assert_near_equal(p.get_val('traj.phase0.timeseries.u')[-1],
                          uf,
                          tolerance=1e-2)

        assert_near_equal(p.get_val('traj.phase0.timeseries.xL')[-1],
                          J,
                          tolerance=5e-4)

    @require_pyoptsparse(optimizer='SLSQP')
    def test_run_brachistochrone_problem(self):
        p = om.Problem(model=om.Group())
        p.driver = om.pyOptSparseDriver(print_results=False)
        p.driver.declare_coloring()
        p.driver.options['optimizer'] = 'SLSQP'

        traj = p.model.add_subsystem('traj', dm.Trajectory())
        phase0 = traj.add_phase('phase0', dm.Phase(ode_class=BrachistochroneODE,
                                                   transcription=dm.Radau(num_segments=2,
                                                                          order=3)))
        phase0.set_time_options(fix_initial=True, fix_duration=False)
        phase0.add_state('x', fix_initial=True, fix_final=False)
        phase0.add_state('y', fix_initial=True, fix_final=False)
        phase0.add_state('v', fix_initial=True, fix_final=False)
        phase0.add_control('theta', continuity=True, rate_continuity=True,
                           units='deg', lower=0.01, upper=179.9)
        phase0.add_parameter('g', units='m/s**2', val=9.80665)

        phase0.add_boundary_constraint('x', loc='final', equals=10)
        phase0.add_boundary_constraint('y', loc='final', equals=5)
        # Minimize time at the end of the phase
        phase0.add_objective('time_phase', loc='final', scaler=10)

        phase0.set_refine_options(refine=True)

        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True)

        phase0.set_time_val(initial=0.0, duration=2.0)
        phase0.set_state_val('x', [0, 10])
        phase0.set_state_val('y', [10, 5])
        phase0.set_state_val('v', [0, 9.9])
        phase0.set_control_val('theta', [5, 100])
        phase0.set_parameter_val('g', 9.80665)

        dm.run_problem(p, simulate=True, simulate_kwargs={'times_per_seg': 100})

        # Assert the results are what we expect.
        sol_db = 'dymos_solution.db'
        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sol_db = p.get_outputs_dir() / sol_db
            sim_db = p.model.traj.sim_prob.get_outputs_dir() / sim_db

        sol_case = om.CaseReader(sol_db).get_case('final')
        sim_case = om.CaseReader(sim_db).get_case('final')
        self.assertTrue(os.path.exists(sol_db))

        sol_t = sol_case.get_val('traj.phase0.timeseries.time')
        sim_t = sim_case.get_val('traj.phase0.timeseries.time')

        assert_almost_equal(sol_case.get_val('traj.phase0.timeseries.time')[-1, ...], 1.8016, decimal=4)
        assert_almost_equal(sim_case.get_val('traj.phase0.timeseries.time')[-1, ...], 1.8016, decimal=2)

        # With two, 3rd-order Radau segments we expect 8 points total.
        self.assertTupleEqual(sol_t.shape, (8, 1))

        # Requested 100 output times per segment, so expect 200 points total
        self.assertTupleEqual(sim_t.shape, (200, 1))

    @require_pyoptsparse(optimizer='SLSQP')
    def test_illegal_simulate_kwargs(self):
        p = om.Problem(model=om.Group())
        p.driver = om.pyOptSparseDriver()
        p.driver.declare_coloring()
        p.driver.options['optimizer'] = 'SLSQP'

        traj = p.model.add_subsystem('traj', dm.Trajectory())
        phase0 = traj.add_phase('phase0', dm.Phase(ode_class=BrachistochroneODE,
                                                   transcription=dm.Radau(num_segments=2,
                                                                          order=3)))
        phase0.set_time_options(fix_initial=True, fix_duration=False)
        phase0.add_state('x', fix_initial=True, fix_final=False)
        phase0.add_state('y', fix_initial=True, fix_final=False)
        phase0.add_state('v', fix_initial=True, fix_final=False)
        phase0.add_control('theta', continuity=True, rate_continuity=True,
                           units='deg', lower=0.01, upper=179.9)
        phase0.add_parameter('g', units='m/s**2', val=9.80665)

        phase0.add_boundary_constraint('x', loc='final', equals=10)
        phase0.add_boundary_constraint('y', loc='final', equals=5)
        # Minimize time at the end of the phase
        phase0.add_objective('time_phase', loc='final', scaler=10)

        phase0.set_refine_options(refine=True)

        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True)

        phase0.set_time_val(initial=0.0, duration=2.0)
        phase0.set_state_val('x', [0, 10])
        phase0.set_state_val('y', [10, 5])
        phase0.set_state_val('v', [0, 9.9])
        phase0.set_control_val('theta', [5, 100])
        phase0.set_parameter_val('g', 9.80665)

        with self.assertRaises(ValueError) as e:
            dm.run_problem(p, simulate=True, simulate_kwargs={'record_file': 'my_sim_file.db'})
        self.assertEqual(str(e.exception),
                         'Key "record_file" was found in simulate_kwargs but should instead by provided by '
                         'the argument "simulation_record_file".')

        with self.assertRaises(ValueError) as e:
            dm.run_problem(p, simulate=True, simulate_kwargs={'case_prefix': 'foo'})
        self.assertEqual(str(e.exception),
                         'Key "case_prefix" was found in simulate_kwargs but should instead by provided by '
                         'the argument "case_prefix", not part of the simulate_kwargs dictionary.')

        with self.assertRaises(ValueError) as e:
            dm.run_problem(p, simulate=True, simulate_kwargs={'case_prefix': 'foo'})
        self.assertEqual(str(e.exception),
                         'Key "case_prefix" was found in simulate_kwargs but should instead by provided by '
                         'the argument "case_prefix", not part of the simulate_kwargs dictionary.')

    @require_pyoptsparse(optimizer='SLSQP')
    def test_run_brachistochrone_problem_refine_case_driver_case_prefix(self):
        p = om.Problem(model=om.Group())
        p.driver = om.pyOptSparseDriver()
        p.driver.declare_coloring()
        p.driver.options['optimizer'] = 'SLSQP'

        p.driver.add_recorder(om.SqliteRecorder('brach_driver_rec.db'))

        traj = p.model.add_subsystem('traj', dm.Trajectory())
        phase0 = traj.add_phase('phase0', dm.Phase(ode_class=BrachistochroneODE,
                                                   transcription=dm.Radau(num_segments=5,
                                                                          order=3)))
        phase0.set_time_options(fix_initial=True, fix_duration=False)
        phase0.add_state('x', fix_initial=True, fix_final=False)
        phase0.add_state('y', fix_initial=True, fix_final=False)
        phase0.add_state('v', fix_initial=True, fix_final=False)
        phase0.add_control('theta', continuity=True, rate_continuity=True,
                           units='deg', lower=0.01, upper=179.9)
        phase0.add_parameter('g', units='m/s**2', val=9.80665)

        phase0.add_boundary_constraint('x', loc='final', equals=10)
        phase0.add_boundary_constraint('y', loc='final', equals=5)
        # Minimize time at the end of the phase
        phase0.add_objective('time_phase', loc='final', scaler=10)

        phase0.set_refine_options(refine=True)

        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True)

        phase0.set_time_val(initial=0.0, duration=2.0)
        phase0.set_state_val('x', [0, 10])
        phase0.set_state_val('y', [10, 5])
        phase0.set_state_val('v', [0, 9.9])
        phase0.set_control_val('theta', [5, 100])
        phase0.set_parameter_val('g', 9.80665)

        dm.run_problem(p, refine_iteration_limit=20)

        sol_db = 'dymos_solution.db'
        driver_db = 'brach_driver_rec.db'
        if om_version()[0] > (3, 34, 2):
            sol_db = p.get_outputs_dir() / sol_db
            driver_db = p.get_outputs_dir() / driver_db

        self.assertTrue(os.path.exists(sol_db))
        # Assert the results are what we expect.
        cr = om.CaseReader(sol_db)
        case = cr.get_case('final')
        assert_almost_equal(case.outputs['traj.phase0.timeseries.time'].max(), 1.8016, decimal=4)

        cr_opt = om.CaseReader(driver_db)
        cases = cr_opt.list_cases(source='driver', out_stream=None)

        for case in cases:
            self.assertTrue(case.startswith('hp_') and 'pyOptSparse_SLSQP|' in case, msg=f'Unexpected case: {case}')

    @require_pyoptsparse(optimizer='SLSQP')
    def test_run_brachistochrone_problem_refine_case_prefix(self):
        p = om.Problem(model=om.Group())
        p.driver = om.pyOptSparseDriver()
        p.driver.declare_coloring()
        p.driver.options['optimizer'] = 'SLSQP'

        p.driver.add_recorder(om.SqliteRecorder('brach_driver_rec.db'))
        p.driver.options['debug_print'] = ['desvars']

        traj = p.model.add_subsystem('traj', dm.Trajectory())
        phase0 = traj.add_phase('phase0', dm.Phase(ode_class=BrachistochroneODE,
                                                   transcription=dm.Radau(num_segments=5,
                                                                          order=3)))
        phase0.set_time_options(fix_initial=True, fix_duration=False)
        phase0.add_state('x', fix_initial=True, fix_final=False)
        phase0.add_state('y', fix_initial=True, fix_final=False)
        phase0.add_state('v', fix_initial=True, fix_final=False)
        phase0.add_control('theta', continuity=True, rate_continuity=True,
                           units='deg', lower=0.01, upper=179.9)
        phase0.add_parameter('g', units='m/s**2', val=9.80665)

        phase0.add_boundary_constraint('x', loc='final', equals=10)
        phase0.add_boundary_constraint('y', loc='final', equals=5)
        # Minimize time at the end of the phase
        phase0.add_objective('time_phase', loc='final', scaler=10)

        phase0.set_refine_options(refine=True)

        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True)

        phase0.set_time_val(initial=0.0, duration=2.0)
        phase0.set_state_val('x', [0, 10])
        phase0.set_state_val('y', [10, 5])
        phase0.set_state_val('v', [0, 9.9])
        phase0.set_control_val('theta', [5, 100])
        phase0.set_parameter_val('g', 9.80665)

        dm.run_problem(p, refine_iteration_limit=20, case_prefix='brach_test')

        sol_db = 'dymos_solution.db'
        driver_db = 'brach_driver_rec.db'
        if om_version()[0] > (3, 34, 2):
            sol_db = p.get_outputs_dir() / sol_db
            driver_db = p.get_outputs_dir() / driver_db

        self.assertTrue(os.path.exists(sol_db))
        # Assert the results are what we expect.
        case = om.CaseReader(sol_db).get_case('brach_test_final')
        assert_almost_equal(case.outputs['traj.phase0.timeseries.time'].max(), 1.8016, decimal=4)

        cr_opt = om.CaseReader(driver_db)
        cases = cr_opt.list_cases(source='driver', out_stream=None)

        for case in cases:
            self.assertTrue(case.startswith('brach_test_hp_') and 'pyOptSparse_SLSQP|' in case, msg=f'Unexpected case: {case}')

    @require_pyoptsparse(optimizer='SLSQP')
    def test_run_brachistochrone_vector_states_problem(self):
        p = om.Problem(model=om.Group())
        p.driver = om.pyOptSparseDriver()
        p.driver.declare_coloring()
        p.driver.options['optimizer'] = 'SLSQP'

        phase = dm.Phase(ode_class=BrachistochroneVectorStatesODE,
                         transcription=dm.Radau(num_segments=1, order=3))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('pos', fix_initial=True, fix_final=[True, False])
        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', units='deg', rate_continuity=False, lower=0.01, upper=179.9)

        phase.add_parameter('g', units='m/s**2', opt=False, val=9.80665)

        phase.add_boundary_constraint('pos', loc='final', equals=5, indices=[1])

        # Minimize time at the end of the phase
        phase.add_objective('time', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True, force_alloc_complex=True)

        phase.set_time_val(initial=0, duration=2.0)

        pos0 = [0, 10]
        posf = [10, 5]
        phase.set_state_val('pos', [pos0, posf])
        phase.set_state_val('v', [0, 9.9])
        phase.set_control_val('theta', [5, 100])
        phase.set_parameter_val('g', 9.80665)

        dm.run_problem(p, refine_iteration_limit=5)

        assert_near_equal(p.get_val('phase0.t')[-1], 1.8016, tolerance=1.0E-3)

    @require_pyoptsparse(optimizer='SLSQP')
    def test_run_brachistochrone_problem_with_simulate(self):
        p = om.Problem(model=om.Group())
        p.driver = om.pyOptSparseDriver()
        p.driver.declare_coloring()
        p.driver.options['optimizer'] = 'SLSQP'

        traj = p.model.add_subsystem('traj', dm.Trajectory())
        phase0 = traj.add_phase('phase0', dm.Phase(ode_class=BrachistochroneODE,
                                                   transcription=dm.Radau(num_segments=10,
                                                                          order=3)))
        phase0.set_time_options(fix_initial=True, fix_duration=False)
        phase0.add_state('x', fix_initial=True, fix_final=False)
        phase0.add_state('y', fix_initial=True, fix_final=False)
        phase0.add_state('v', fix_initial=True, fix_final=False)
        phase0.add_control('theta', continuity=True, rate_continuity=True,
                           units='deg', lower=0.01, upper=179.9)
        phase0.add_parameter('g', units='m/s**2', val=9.80665)

        phase0.add_boundary_constraint('x', loc='final', equals=10)
        phase0.add_boundary_constraint('y', loc='final', equals=5)
        # Minimize time at the end of the phase
        phase0.add_objective('time_phase', loc='final', scaler=10)

        phase0.set_refine_options(refine=True)

        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True)

        phase0.set_time_val(initial=0.0, duration=2.0)
        phase0.set_state_val('x', [0, 10])
        phase0.set_state_val('y', [10, 5])
        phase0.set_state_val('v', [0, 9.9])
        phase0.set_control_val('theta', [5, 100])
        phase0.set_parameter_val('g', 9.80665)

        dm.run_problem(p, simulate=True)

        sol_db = 'dymos_solution.db'
        if om_version()[0] > (3, 34, 2):
            sol_db = p.get_outputs_dir() / sol_db

        self.assertTrue(os.path.exists(sol_db))

        # Assert the results are what we expect.
        case = om.CaseReader(sol_db).get_case('final')
        assert_almost_equal(case.outputs['traj.phase0.timeseries.time'].max(), 1.8016, decimal=4)

    def test_restart_from_file(self):
        from dymos.examples.vanderpol.vanderpol_dymos import vanderpol
        from dymos.run_problem import run_problem
        from openmdao.components.interp_util.interp import InterpND
        from numpy.testing import assert_almost_equal

        # Create the Dymos problem instance
        p = vanderpol(transcription='gauss-lobatto', num_segments=75)

        # Run the problem (simulate only)
        dm.run_problem(p, run_driver=False, simulate=True)

        # Run the model
        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sim_db = p.model.traj.sim_prob.get_outputs_dir() / sim_db

        # create a new problem for restart to simulate a different command line execution
        q = vanderpol(transcription='gauss-lobatto', num_segments=75)

        run_problem(q, run_driver=False, simulate=False, restart=sim_db)

        sol_db2 = 'dymos_solution.db'
        if om_version()[0] > (3, 34, 2):
            sol_db2 = q.get_outputs_dir() / sol_db2
        s = om.CaseReader(sol_db2).get_case('final')

        # get_val returns data for duplicate time points; remove them before interpolating
        tq = q.get_val('traj.phase0.timeseries.time')[:, 0]
        nodup = np.insert(tq[1:] != tq[:-1], 0, True)
        tq = tq[nodup]
        x1q = q.get_val('traj.phase0.timeseries.x1')[:, 0][nodup]
        x0q = q.get_val('traj.phase0.timeseries.x0')[:, 0][nodup]
        uq = q.get_val('traj.phase0.timeseries.u')[:, 0][nodup]

        ts = s.get_val('traj.phase0.timeseries.time')[:, 0]
        nodup = np.insert(ts[1:] != ts[:-1], 0, True)
        ts = ts[nodup]
        x1s = s.get_val('traj.phase0.timeseries.x1')[:, 0][nodup]
        x0s = s.get_val('traj.phase0.timeseries.x0')[:, 0][nodup]
        us = s.get_val('traj.phase0.timeseries.u')[:, 0][nodup]

        # create interpolation functions so that values can be looked up at matching time points
        fx1s = InterpND('cubic', ts, x1s).interpolate
        fx0s = InterpND('cubic', ts, x0s).interpolate
        fus = InterpND('cubic', ts, us).interpolate

        assert_almost_equal(x1q, fx1s(tq), decimal=2)
        assert_almost_equal(x0q, fx0s(tq), decimal=2)
        assert_almost_equal(uq, fus(tq), decimal=5)

    def test_restart_from_case(self):
        from dymos.examples.vanderpol.vanderpol_dymos import vanderpol
        from dymos.run_problem import run_problem
        from openmdao.components.interp_util.interp import InterpND
        from numpy.testing import assert_almost_equal

        # Create the Dymos problem instance
        p = vanderpol(transcription='gauss-lobatto', num_segments=75)

        # Run the problem (simulate only)
        dm.run_problem(p, run_driver=False, simulate=True)

        # create a new problem for restart to simulate a different command line execution
        q = vanderpol(transcription='gauss-lobatto', num_segments=75)

        # Run the model
        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sim_db = p.model.traj.sim_prob.get_outputs_dir() / sim_db

        run_problem(q, run_driver=False, simulate=False, restart=sim_db)

        sol_db2 = 'dymos_solution.db'
        if om_version()[0] > (3, 34, 2):
            sol_db2 = q.get_outputs_dir() / sol_db2
        s = om.CaseReader(sol_db2).get_case('final')

        # get_val returns data for duplicate time points; remove them before interpolating
        tq = q.get_val('traj.phase0.timeseries.time')[:, 0]
        nodup = np.insert(tq[1:] != tq[:-1], 0, True)
        tq = tq[nodup]
        x1q = q.get_val('traj.phase0.timeseries.x1')[:, 0][nodup]
        x0q = q.get_val('traj.phase0.timeseries.x0')[:, 0][nodup]
        uq = q.get_val('traj.phase0.timeseries.u')[:, 0][nodup]

        ts = s.get_val('traj.phase0.timeseries.time')[:, 0]
        nodup = np.insert(ts[1:] != ts[:-1], 0, True)
        ts = ts[nodup]
        x1s = s.get_val('traj.phase0.timeseries.x1')[:, 0][nodup]
        x0s = s.get_val('traj.phase0.timeseries.x0')[:, 0][nodup]
        us = s.get_val('traj.phase0.timeseries.u')[:, 0][nodup]

        # create interpolation functions so that values can be looked up at matching time points
        fx1s = InterpND('cubic', ts, x1s).interpolate
        fx0s = InterpND('cubic', ts, x0s).interpolate
        fus = InterpND('cubic', ts, us).interpolate

        assert_almost_equal(x1q, fx1s(tq), decimal=2)
        assert_almost_equal(x0q, fx0s(tq), decimal=2)
        assert_almost_equal(uq, fus(tq), decimal=5)


@use_tempdirs
@require_pyoptsparse(optimizer='SLSQP')
class TestRunProblemPlotting(unittest.TestCase):
    def setUp(self):
        p = om.Problem(model=om.Group())
        p.driver = om.pyOptSparseDriver()
        p.driver.declare_coloring()
        p.driver.options['optimizer'] = 'SLSQP'

        traj = p.model.add_subsystem('traj', dm.Trajectory())
        phase0 = traj.add_phase('phase0', dm.Phase(ode_class=BrachistochroneODE,
                                                   transcription=dm.Radau(num_segments=10,
                                                                          order=3)))
        phase0.set_time_options(fix_initial=True, fix_duration=True)
        phase0.add_state('x', fix_initial=True, fix_final=False)
        phase0.add_state('y', fix_initial=True, fix_final=False)
        phase0.add_state('v', fix_initial=True, fix_final=False)
        phase0.add_control('theta', continuity=True, rate_continuity=True,
                           units='deg', lower=0.01, upper=179.9)
        phase0.add_parameter('g', units='m/s**2', val=9.80665)

        phase0.add_boundary_constraint('x', loc='final', equals=10)
        phase0.add_boundary_constraint('y', loc='final', equals=5)
        # Minimize time at the end of the phase
        phase0.add_objective('time_phase', loc='final', scaler=10)

        phase0.set_refine_options(refine=True)

        phase0.timeseries_options['include_state_rates'] = True
        phase0.timeseries_options['include_control_rates'] = True
        phase0.timeseries_options['include_t_phase'] = True

        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True)

        phase0.set_time_val(initial=0.0, duration=2.0)
        phase0.set_state_val('x', [0, 10])
        phase0.set_state_val('y', [10, 5])
        phase0.set_state_val('v', [0, 9.9])
        phase0.set_control_val('theta', [5, 100])
        phase0.set_parameter_val('g', 9.80665)

        self.p = p

    @unittest.skipIf(matplotlib is None, "This test requires matplotlib")
    def test_run_brachistochrone_problem_make_plots(self):
        plots_cache = dm.options['plots']
        dm.options['plots'] = 'matplotlib'
        dm.run_problem(self.p, make_plots=True)
        plot_dir = pathlib.Path(_get_reports_dir(self.p)).joinpath('plots')

        for varname in ['time_phase', 'x', 'xdot', 'y',
                        'ydot', 'v', 'vdot', 'theta', 'theta_rate', 'theta_rate2']:
            self.assertTrue(plot_dir.joinpath(f'{varname}.png').exists(),
                            msg=f'{varname}.png' + ' does not exist.')
        dm.options['plots'] = plots_cache

    @unittest.skipIf(matplotlib is None, "This test requires matplotlib")
    def test_run_brachistochrone_problem_make_plots_set_plot_dir(self):
        _cache = dm.options['plots']
        dm.options['plots'] = 'matplotlib'

        dm.run_problem(self.p, make_plots=True, plot_dir="test_plot_dir")

        plot_dir = pathlib.Path(_get_reports_dir(self.p))
        for varname in ['x', 'y', 'v', 'theta']:
            plotfile = plot_dir.joinpath('test_plot_dir', f'{varname}.png')
            self.assertTrue(str(plotfile) + ' does not exist.')

        dm.options['plots'] = _cache

    def test_run_brachistochrone_problem_do_not_make_plots(self):
        dm.run_problem(self.p, make_plots=False)
        plot_dir = pathlib.Path(_get_reports_dir(self.p)).joinpath('plots')

        for varname in ['time_phase', 'x', 'xdot', 'y', 'ydot', 'v',
                        'vdot', 'theta', 'theta_rate']:
            plotfile = plot_dir.joinpath(f'{varname}.png')
            self.assertFalse(plotfile.exists(), msg=f'Unexpectedly found plot file {plotfile}')

    def test_run_brachistochrone_problem_set_simulation_record_file(self):
        sim_db = 'simulation_record_file.db'
        dm.run_problem(self.p, simulate=True, simulation_record_file=sim_db)

        if om_version()[0] > (3, 34, 2):
            sim_db = self.p.model.traj.sim_prob.get_outputs_dir() / sim_db

        self.assertTrue(os.path.exists(sim_db))

    def test_run_brachistochrone_problem_set_solution_record_file(self):
        sol_db = 'solution_record_file.db'
        dm.run_problem(self.p, solution_record_file=sol_db)

        if om_version()[0] > (3, 34, 2):
            sol_db = self.p.get_outputs_dir() / sol_db

        self.assertTrue(os.path.exists(sol_db))

    @unittest.skipIf(matplotlib is None, "This test requires matplotlib")
    def test_run_brachistochrone_problem_plot_simulation(self):
        plots_cache = dm.options['plots']
        dm.options['plots'] = 'matplotlib'

        dm.run_problem(self.p, make_plots=True, simulate=True)
        plot_dir = pathlib.Path(_get_reports_dir(self.p)).joinpath('plots')

        for varname in ['time_phase', 'x', 'xdot', 'y',
                        'ydot', 'v', 'vdot', 'theta', 'theta_rate', 'theta_rate2']:
            plotfile = plot_dir.joinpath(f'{varname}.png')
            self.assertTrue(plotfile.exists(), msg=f'plot file {plotfile} does not exist!')
        dm.options['plots'] = plots_cache

    @unittest.skipIf(matplotlib is None, "This test requires matplotlib")
    def test_run_brachistochrone_problem_plot_no_simulation_record_file_given(self):
        plots_cache = dm.options['plots']
        dm.options['plots'] = 'matplotlib'

        dm.run_problem(self.p, make_plots=True, simulate=True)
        plot_dir = pathlib.Path(_get_reports_dir(self.p)).joinpath('plots')

        for varname in ['time_phase', 'x', 'xdot', 'y',
                        'ydot', 'v', 'vdot', 'theta', 'theta_rate', 'theta_rate2']:
            plotfile = plot_dir.joinpath(f'{varname.replace(":", "_")}.png')
            self.assertTrue(plotfile.exists(), msg=f'plot file {plotfile} does not exist!')

        dm.options['plots'] = plots_cache


@use_tempdirs
class TestSimulateArrayParam(unittest.TestCase):

    def test_simulate_array_param(self):
        #
        # Initialize the Problem and the optimization driver
        #
        p = om.Problem(model=om.Group())
        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        # dummy array of data
        indeps = p.model.add_subsystem('indeps', om.IndepVarComp(), promotes=['*'])
        indeps.add_output('array', np.linspace(1, 10, 10), units=None)

        #
        # Create a trajectory and add a phase to it
        #
        traj = p.model.add_subsystem('traj', dm.Trajectory())

        phase = traj.add_phase('phase0',
                               dm.Phase(ode_class=BrachistochroneODE,
                                        transcription=dm.GaussLobatto(num_segments=10)))

        #
        # Set the variables
        #
        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=True, rate_source='xdot')

        phase.add_state('y', fix_initial=True, fix_final=True, rate_source='ydot')

        phase.add_state('v', fix_initial=True, fix_final=False, rate_source='vdot')

        phase.add_control('theta', continuity=True, rate_continuity=True,
                          units='deg', lower=0.01, upper=179.9)

        phase.add_parameter('g', units='m/s**2', val=9.80665)

        # add dummy array as a parameter and connect it
        phase.add_parameter('array', units=None, shape=(10,), static_target=True)
        p.model.connect('array', 'traj.phase0.parameters:array')

        #
        # Minimize time at the end of the phase
        #
        phase.add_objective('time', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()

        #
        # Setup the Problem
        #
        p.setup()

        #
        # Set the initial values
        #
        phase.set_time_val(initial=0.0, duration=2.0)
        phase.set_state_val('x', [0, 10])
        phase.set_state_val('y', [10, 5])
        phase.set_state_val('v', [0, 9.9])
        phase.set_control_val('theta', [5, 100])
        phase.set_parameter_val('g', 9.80665)

        #
        # Solve for the optimal trajectory
        #
        dm.run_problem(p, simulate=True)

        # Test the results
        sol_db = 'dymos_solution.db'
        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sol_db = p.get_outputs_dir() / sol_db
            sim_db = p.model.traj.sim_prob.get_outputs_dir() / sim_db

        sol_results = om.CaseReader(sol_db).get_case('final')
        sim_results = om.CaseReader(sim_db).get_case('final')

        sol = sol_results.get_val('traj.phase0.parameter_vals:array')
        sim = sim_results.get_val('traj.phase0.parameter_vals:array')

        assert_near_equal(sol - sim, np.zeros_like(sol))


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import numpy as np
from openmdao.api import ExplicitComponent


class FlightDynamics(ExplicitComponent):
    """
    Defines the flight dynamics for the shuttle reentry problem.

    References
    ----------
    .. [1] Betts, John T., Practical Methods for Optimal Control and Estimation Using Nonlinear
           Programming, p. 247, 2010.
    """

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        self.add_input('beta', val=np.ones(nn), desc='bank angle', units='rad')
        self.add_input('gamma', val=np.ones(nn), desc='flight path angle', units='rad')
        self.add_input('h', val=np.ones(nn), desc='altitude of shuttle', units='ft')
        self.add_input('psi', val=np.ones(nn), desc='azimuthal angle', units='rad')
        self.add_input('theta', val=np.ones(nn), desc='latitude', units='rad')
        self.add_input('v', val=np.ones(nn), desc='velocity of shuttle', units='ft/s')
        self.add_input('lift', val=np.ones(nn), desc='lift on shuttle', units='lb')
        self.add_input('drag', val=np.ones(nn), desc='drag on shuttle', units='lb')

        self.add_output('hdot', val=np.ones(nn), desc='rate of change of altitude',
                        units='ft/s')
        self.add_output('gammadot', val=np.ones(nn),
                        desc='rate of change of flight path angle', units='rad/s')
        self.add_output('phidot', val=np.ones(nn), desc='rate of change of longitude',
                        units='rad/s')
        self.add_output('psidot', val=np.ones(nn), desc='rate of change of azimuthal angle',
                        units='rad/s')
        self.add_output('thetadot', val=np.ones(nn), desc='rate of change of latitude',
                        units='rad/s')
        self.add_output('vdot', val=np.ones(nn), desc='rate of change of velocity',
                        units='ft/s**2')

        partial_range = np.arange(nn, dtype=int)

        self.declare_partials('hdot', 'v', rows=partial_range, cols=partial_range)
        self.declare_partials('hdot', 'gamma', rows=partial_range, cols=partial_range)

        self.declare_partials('gammadot', 'lift', rows=partial_range, cols=partial_range)
        self.declare_partials('gammadot', 'h', rows=partial_range, cols=partial_range)
        self.declare_partials('gammadot', 'beta', rows=partial_range, cols=partial_range)
        self.declare_partials('gammadot', 'gamma', rows=partial_range, cols=partial_range)
        self.declare_partials('gammadot', 'v', rows=partial_range, cols=partial_range)

        self.declare_partials('phidot', 'v', rows=partial_range, cols=partial_range)
        self.declare_partials('phidot', 'h', rows=partial_range, cols=partial_range)
        self.declare_partials('phidot', 'gamma', rows=partial_range, cols=partial_range)
        self.declare_partials('phidot', 'psi', rows=partial_range, cols=partial_range)
        self.declare_partials('phidot', 'theta', rows=partial_range, cols=partial_range)

        self.declare_partials('psidot', 'v', rows=partial_range, cols=partial_range)
        self.declare_partials('psidot', 'gamma', rows=partial_range, cols=partial_range)
        self.declare_partials('psidot', 'h', rows=partial_range, cols=partial_range)
        self.declare_partials('psidot', 'beta', rows=partial_range, cols=partial_range)
        self.declare_partials('psidot', 'theta', rows=partial_range, cols=partial_range)
        self.declare_partials('psidot', 'psi', rows=partial_range, cols=partial_range)
        self.declare_partials('psidot', 'lift', rows=partial_range, cols=partial_range)

        self.declare_partials('thetadot', 'v', rows=partial_range, cols=partial_range)
        self.declare_partials('thetadot', 'h', rows=partial_range, cols=partial_range)
        self.declare_partials('thetadot', 'gamma', rows=partial_range, cols=partial_range)
        self.declare_partials('thetadot', 'psi', rows=partial_range, cols=partial_range)

        self.declare_partials('vdot', 'drag', rows=partial_range, cols=partial_range)
        self.declare_partials('vdot', 'gamma', rows=partial_range, cols=partial_range)
        self.declare_partials('vdot', 'h', rows=partial_range, cols=partial_range)

    def compute(self, inputs, outputs):
        v = inputs['v']
        gamma = inputs['gamma']
        theta = inputs['theta']
        lift = inputs['lift']
        drag = inputs['drag']
        h = inputs['h']
        beta = inputs['beta']
        psi = inputs['psi']
        g_0 = 32.174
        w = 203000
        R_e = 20902900
        mu = .14076539e17
        s_beta = np.sin(beta)
        c_beta = np.cos(beta)
        s_gamma = np.sin(gamma)
        c_gamma = np.cos(gamma)
        s_psi = np.sin(psi)
        c_psi = np.cos(psi)
        c_theta = np.cos(theta)
        s_theta = np.sin(theta)
        r = R_e + h
        m = w / g_0
        g = mu / r ** 2

        outputs['hdot'] = v * s_gamma
        outputs['gammadot'] = lift / (m * v) * c_beta + c_gamma * (v / r - g / v)
        outputs['phidot'] = v / r * c_gamma * s_psi / c_theta
        outputs['psidot'] = lift * s_beta / (m * v * c_gamma) + \
            v * c_gamma * s_psi * s_theta / (r * c_theta)
        outputs['thetadot'] = c_gamma * c_psi * v / r
        outputs['vdot'] = -drag / m - g * s_gamma

    def compute_partials(self, inputs, J):
        v = inputs['v']
        gamma = inputs['gamma']
        theta = inputs['theta']
        lift = inputs['lift']
        h = inputs['h']
        beta = inputs['beta']
        psi = inputs['psi']
        g_0 = 32.174
        w = 203000
        R_e = 20902900
        mu = .14076539e17
        s_beta = np.sin(beta)
        c_beta = np.cos(beta)
        s_gamma = np.sin(gamma)
        c_gamma = np.cos(gamma)
        s_psi = np.sin(psi)
        c_psi = np.cos(psi)
        c_theta = np.cos(theta)
        s_theta = np.sin(theta)
        r = R_e + h
        m = w / g_0
        g = mu / r ** 2

        J['hdot', 'v'] = s_gamma
        J['hdot', 'gamma'] = v * c_gamma

        J['gammadot', 'lift'] = c_beta / (m * v)
        J['gammadot', 'h'] = c_gamma * (-v / r ** 2 + 2 * mu / (r ** 3 * v))
        J['gammadot', 'beta'] = -lift / (m * v) * s_beta
        J['gammadot', 'gamma'] = -s_gamma * (v / r - g / v)
        J['gammadot', 'v'] = -lift / (m * v ** 2) * c_beta + c_gamma * (1 / r + g / v ** 2)

        J['phidot', 'v'] = c_gamma * s_psi / (c_theta * r)
        J['phidot', 'h'] = -v / r ** 2 * c_gamma * s_psi / c_theta
        J['phidot', 'gamma'] = -v / r * s_gamma * s_psi / c_theta
        J['phidot', 'psi'] = v / r * c_gamma * c_psi / c_theta
        J['phidot', 'theta'] = v / r * c_gamma * s_psi / (c_theta ** 2) * s_theta

        J['psidot', 'v'] = -lift * s_beta / (m * c_gamma * v ** 2) + \
            c_gamma * s_psi * s_theta / (r * c_theta)
        J['psidot', 'gamma'] = lift * s_beta / (m * v * c_gamma ** 2) * s_gamma - \
            v * s_gamma * s_psi * s_theta / (r * c_theta)
        J['psidot', 'h'] = -v * c_gamma * s_psi * s_theta / (c_theta * r ** 2)
        J['psidot', 'beta'] = lift * c_beta / (m * v * c_gamma)
        J['psidot', 'theta'] = v * c_gamma * s_psi / (r * c_theta ** 2)
        J['psidot', 'psi'] = v * c_gamma * c_psi * s_theta / (r * c_theta)
        J['psidot', 'lift'] = s_beta / (m * v * c_gamma)

        J['thetadot', 'v'] = c_gamma * c_psi / r
        J['thetadot', 'h'] = -v / r ** 2 * c_gamma * c_psi
        J['thetadot', 'gamma'] = -v / r * s_gamma * c_psi
        J['thetadot', 'psi'] = -v / r * c_gamma * s_psi

        J['vdot', 'h'] = 2 * s_gamma * mu / r ** 3
        J['vdot', 'drag'] = -1 / m
        J['vdot', 'gamma'] = -g * c_gamma



"""
Generic utilities for use by the grid refinement schemes.
"""
import numpy as np
from scipy.linalg import block_diag

import openmdao.api as om

from ..transcriptions import GaussLobatto, Radau
from ..utils.lagrange import lagrange_matrices
from ..utils.misc import get_rate_units
from ..utils.introspection import get_targets

from .grid_refinement_ode_system import GridRefinementODESystem


def interpolation_lagrange_matrix(old_grid, new_grid):
    """
    Evaluate lagrange matrix to interpolate state and control values from the solved grid onto the new grid.

    Parameters
    ----------
    old_grid : <GridData>
        GridData object representing the grid on which the problem has been solved.
    new_grid : <GridData>
        GridData object representing the new, higher-order grid.

    Returns
    -------
    ndarray
        The lagrange interpolation matrix.
    """
    L_blocks = []
    D_blocks = []

    for iseg in range(old_grid.num_segments):
        i1, i2 = old_grid.subset_segment_indices['all'][iseg, :]
        indices = old_grid.subset_node_indices['all'][i1:i2]
        nodes_given = old_grid.node_stau[indices]

        i1, i2 = new_grid.subset_segment_indices['all'][iseg, :]
        indices = new_grid.subset_node_indices['all'][i1:i2]
        nodes_eval = new_grid.node_stau[indices]

        L_block, D_block = lagrange_matrices(nodes_given, nodes_eval)

        L_blocks.append(L_block)
        D_blocks.append(D_block)

    L = block_diag(*L_blocks)
    D = block_diag(*D_blocks)

    return L, D


def integration_matrix(grid):
    """
    Evaluate the Integration matrix of the given grid.

    Parameters
    ----------
    grid : <GridData>
        GridData object containing the grid where the integration matrix will be evaluated.

    Returns
    -------
    ndarray
        The integration matrix used to propagate initial states over segments.
    """
    I_blocks = []

    for iseg in range(grid.num_segments):
        i1, i2 = grid.subset_segment_indices['all'][iseg, :]
        indices = grid.subset_node_indices['all'][i1:i2]
        nodes_given = grid.node_stau[indices]

        i1, i2 = grid.subset_segment_indices['all'][iseg, :]
        indices = grid.subset_node_indices['all'][i1:i2]
        nodes_eval = grid.node_stau[indices][1:]

        _, D_block = lagrange_matrices(nodes_given, nodes_eval)
        I_block = np.linalg.inv(D_block[:, 1:])
        I_blocks.append(I_block)

    return block_diag(*I_blocks)


def eval_ode_on_grid(phase, transcription):
    """
    Evaluate the ODE from the given phase at all nodes of the given transcription.

    Parameters
    ----------
    phase : Phase
        A Phase object which has been executed and whose ODE is to be run at all nodes
        of the given transcription.
    transcription : Radau or GaussLobatto transcription instance
        The transcription object at which to execute the ODE of the given phase at all nodes.

    Returns
    -------
    dict of (str: ndarray)
        A dictionary of the state values from the phase interpolated to the new transcription.
    dict of (str: ndarray)
        A dictionary of the control values from the phase interpolated to the new transcription.
    dict of (str: ndarray)
        A dictionary of the state rates computed in the phase's ODE at the new transcription points.
    """
    x = {}
    u = {}
    u_rate = {}
    u_rate2 = {}
    p = {}
    param = {}
    f = {}

    # Build the interpolation matrix which interpolates from all nodes on the old grid to
    # all nodes on the new grid.
    grid_data = transcription.grid_data
    L, _ = interpolation_lagrange_matrix(old_grid=phase.options['transcription'].grid_data,
                                         new_grid=grid_data)

    # Create a new problem for the grid_refinement
    # For this test, use the same grid as the original problem.
    p_refine = om.Problem(model=om.Group())
    grid_refinement_system = GridRefinementODESystem(grid_data=grid_data,
                                                     time=phase.time_options,
                                                     states=phase.state_options,
                                                     controls=phase.control_options,
                                                     parameters=phase.parameter_options,
                                                     ode_class=phase.options['ode_class'],
                                                     ode_init_kwargs=phase.options[
                                                         'ode_init_kwargs'],
                                                     calc_exprs=phase._calc_exprs)
    p_refine.model.add_subsystem('grid_refinement_system', grid_refinement_system, promotes=['*'])
    p_refine.setup()

    # Set the values in the refinement problem using the outputs from the first
    ode = p_refine.model.grid_refinement_system.ode

    t_name = phase.time_options['name']

    t_prev = phase.get_val(f'timeseries.{t_name}', units=phase.time_options['units'])
    t_phase_prev = t_prev - t_prev[0]
    t_initial = np.repeat(t_prev[0, 0], repeats=transcription.grid_data.num_nodes, axis=0)
    t_duration = np.repeat(t_prev[-1, 0], repeats=transcription.grid_data.num_nodes, axis=0)
    t = np.dot(L, t_prev)
    t_phase = np.dot(L, t_phase_prev)
    targets = get_targets(ode, 'time', phase.time_options['targets'])
    t_phase_targets = get_targets(ode, 't_phase', phase.time_options['time_phase_targets'])
    t_initial_targets = get_targets(ode, 't_initial', phase.time_options['t_initial_targets'])
    t_duration_targets = get_targets(ode, 't_duration', phase.time_options['t_duration_targets'])
    if targets:
        p_refine.set_val('time', t)
    if t_phase_targets:
        p_refine.set_val('t_phase', t_phase)
    if t_initial_targets:
        p_refine.set_val('t_initial', t_initial)
    if t_duration_targets:
        p_refine.set_val('t_duration', t_duration)

    state_prefix = 'states:' if phase.timeseries_options['use_prefix'] else ''
    control_prefix = 'controls:' if phase.timeseries_options['use_prefix'] else ''

    for name, options in phase.state_options.items():
        x_prev = phase.get_val(f'timeseries.{state_prefix}{name}', units=options['units'])
        x[name] = np.dot(L, x_prev)
        targets = get_targets(ode, name, options['targets'])
        if targets:
            p_refine.set_val(f'states:{name}', x[name])

    for name, options in phase.control_options.items():
        targets = get_targets(ode, name, options['targets'])
        rate_targets = get_targets(ode, f'{name}_rate', options['rate_targets'])
        rate2_targets = get_targets(ode, f'{name}_rate2', options['rate2_targets'])

        u_prev = phase.get_val(f'timeseries.{control_prefix}{name}', units=options['units'])
        u[name] = np.dot(L, u_prev)
        if targets:
            p_refine.set_val(f'controls:{name}', u[name])

        if phase.timeseries_options['include_control_rates']:
            if rate_targets:
                u_rate_prev = phase.get_val(f'timeseries.control_rates:{name}_rate')
                u_rate[name] = np.dot(L, u_rate_prev)
                p_refine.set_val(f'control_rates:{name}_rate', u_rate[name])

            if rate2_targets:
                u_rate2_prev = phase.get_val(f'timeseries.control_rates:{name}_rate2')
                u_rate2[name] = np.dot(L, u_rate2_prev)
                p_refine.set_val(f'control_rates:{name}_rate2', u_rate2[name])

    # Configure the parameters
    for name, options in phase.parameter_options.items():
        targets = get_targets(ode, name, options['targets'])
        # The value of the parameter at one node
        param[name] = phase.get_val(f'parameter_vals:{name}', units=options['units'])[0, ...]
        if targets:
            p_refine.set_val(f'parameters:{name}', param[name], units=options['units'])

    # Execute the model
    p_refine.run_model()

    # Assign the state rates on the new grid to f
    for name, options in phase.state_options.items():
        rate_units = get_rate_units(options['units'], phase.time_options['units'])
        rate_source = options['rate_source']
        rate_source_class = phase.classify_var(rate_source)
        if rate_source_class in {'time'}:
            src_units = phase.time_options['units']
            f[name] = om.convert_units(t, src_units, rate_units)
        elif rate_source_class in {'time_phase'}:
            src_units = phase.time_options['units']
            f[name] = om.convert_units(t_phase, src_units, rate_units)
        elif rate_source_class in {'state'}:
            src_units = phase.state_options[rate_source]['units']
            f[name] = om.convert_units(x[rate_source], src_units, rate_units)
        elif rate_source_class in {'control'}:
            src_units = phase.control_options[rate_source]['units']
            f[name] = om.convert_units(u[rate_source], src_units, rate_units)
        elif rate_source_class in {'control_rate'}:
            u_name = rate_source[:-5]
            u_units = phase.control_options[u_name]['units']
            src_units = get_rate_units(u_units, phase.time_options['units'], deriv=1)
            f[name] = om.convert_units(u_rate[rate_source], src_units, rate_units)
        elif rate_source_class in {'control_rate2'}:
            u_name = rate_source[:-6]
            u_units = phase.control_options[u_name]['units']
            src_units = get_rate_units(u_units, phase.time_options['units'], deriv=2)
            f[name] = om.convert_units(u_rate2[rate_source], src_units, rate_units)
        elif rate_source_class in {'parameter'}:
            src_units = phase.parameter_options[rate_source]['units']
            shape = phase.parameter_options[rate_source]['shape']
            f[name] = om.convert_units(param[rate_source], src_units, rate_units)
            f[name] = np.broadcast_to(f[name], (grid_data.num_nodes,) + shape)
        elif rate_source_class in {'ode'}:
            f[name] = p_refine.get_val(f'ode.{rate_source}', units=rate_units)

        if len(f[name].shape) == 1:
            f[name] = np.reshape(f[name], (f[name].shape[0], 1))

    return x, u, p, f


def compute_state_quadratures(x_hat, f_hat, t_duration, transcription):
    """
    Compute the integral of the given states at each node in the given transcription.

    This estimation of the states is computed with a quadrature.

    Parameters
    ----------
    x_hat : dict of (str: float)
        The interpolated state values at the nodes for each state.
    f_hat : dict of (str: float)
        State rates computed using the interpolated state values at each node.
    t_duration : float
        The time duration of the phase.
    transcription : Radau or GaussLobatto
        The transcription instance used to compute f_hat.

    Returns
    -------
    dict
        A dictionary keyed by state name containing the estimated state values at each node,
        computed using a quadrature.
    """
    x_prime = {}
    gd = transcription.grid_data

    # Build the integration matrix which integrates state values at all nodes on the new grid.
    I = integration_matrix(gd)  # noqa: E741, allow ambiguous variable name `I`

    left_end_idxs = gd.subset_node_indices['segment_ends'][0::2]
    all_idxs = gd.subset_node_indices['all']
    not_left_end_idxs = np.array(sorted(set(all_idxs).difference(left_end_idxs)))

    dt_dstau = np.atleast_2d(0.5 * t_duration * gd.node_dptau_dstau[not_left_end_idxs]).T

    if x_hat.keys() != f_hat.keys():
        raise ValueError('x_hat and f_hat don\'t contain the same states.\n'
                         f'x_hat states are: {list(x_hat.keys())}\n'
                         f'f_hat states are: {list(f_hat.keys())}')

    for state_name in x_hat:
        x_prime[state_name] = np.zeros_like(x_hat[state_name])
        x_prime[state_name][left_end_idxs, ...] = x_hat[state_name][left_end_idxs, ...]
        nnps = np.array(gd.subset_num_nodes_per_segment['all']) - 1
        left_end_idxs_repeated = np.repeat(left_end_idxs, nnps)
        x_prime[state_name][not_left_end_idxs, ...] = \
            x_hat[state_name][left_end_idxs_repeated, ...] \
            + dt_dstau * np.dot(I, f_hat[state_name][not_left_end_idxs, ...])

    return x_prime


def check_error(phases):
    """
    Compute the error in every solved segment of the given phases.

    Parameters
    ----------
    phases : dict
        Dict of phase paths and phases.

    Returns
    -------
    dict
        Indicator for which segments of each phase require grid refinement.
    """
    refine_results = {}

    for phase_path, phase in phases.items():
        refine_results[phase_path] = {}

        # Save the original grid to the refine results
        tx = phase.options['transcription']
        gd = tx.grid_data
        numseg = gd.num_segments

        refine_results[phase_path]['num_segments'] = numseg
        refine_results[phase_path]['order'] = gd.transcription_order
        refine_results[phase_path]['segment_ends'] = gd.segment_ends
        refine_results[phase_path]['need_refinement'] = np.zeros(numseg, dtype=bool)
        refine_results[phase_path]['max_rel_error'] = np.zeros(numseg, dtype=float)  # Eq. 21
        refine_results[phase_path]['error_state'] = ['' for _ in phase.state_options]

        # Instantiate a new phase as a copy of the old one, but first up the transcription order
        # by 1 for Radau and by 2 for Gauss-Lobatto
        new_num_segments = tx.options['num_segments']
        new_segment_ends = tx.options['segment_ends']
        new_compressed = tx.options['compressed']
        if isinstance(tx, GaussLobatto):
            new_order = tx.options['order'] + 2
            new_tx = GaussLobatto(num_segments=new_num_segments, order=new_order,
                                  segment_ends=new_segment_ends, compressed=new_compressed)
        elif isinstance(tx, Radau):
            new_order = tx.options['order'] + 1
            new_tx = Radau(num_segments=new_num_segments, order=new_order,
                           segment_ends=new_segment_ends, compressed=new_compressed)
        else:
            # Only refine GuassLobatto or Radau transcription phases
            continue

        # Let x be the interpolated states on the new transcription
        # Let f by the evaluated state rates given the interpolation of the states and controls
        # onto the new grid.
        x, _, _, f = eval_ode_on_grid(phase=phase, transcription=new_tx)

        # x_hat is the state value at each node computed using a quadrature
        # from the initial state value in each segment and the computed state rates
        # at each node in the new transcription.
        x_hat = compute_state_quadratures(x, f, phase.get_val('t_duration'), new_tx)

        E = {}  # The absolute error computed in each state at each node (Eq. 20 pt 1)
        e = {}  # The relative error computed in each state at each node (Eq. 20 pt 2)

        for state_name in phase.state_options:
            E[state_name] = np.abs(x_hat[state_name] - x[state_name])  # Equation 20.1
            e[state_name] = np.zeros_like(E[state_name])               # Equation 20.2
            for k in range(numseg):
                i1, i2 = new_tx.grid_data.subset_segment_indices['all'][k, :]
                k_idxs = new_tx.grid_data.subset_node_indices['all'][i1:i2]
                e[state_name][k_idxs, ...] = E[state_name][k_idxs] \
                    / (1.0 + np.max(np.abs(x[state_name][k_idxs])))
                if np.any(np.max(e[state_name][k_idxs]) > refine_results[phase_path]['max_rel_error'][k]):
                    refine_results[phase_path]['max_rel_error'][k] = np.max(e[state_name][k_idxs])
                    refine_results[phase_path]['error_state'] = state_name
                    if refine_results[phase_path]['max_rel_error'][k] > phase.refine_options['tolerance']:
                        refine_results[phase_path]['need_refinement'][k] = True

    return refine_results



import numpy as np
import openmdao.api as om


class DynamicPressureComp(om.ExplicitComponent):
    """ Compute the dynamic pressure based on the velocity and the atmospheric density. """
    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        # Inputs
        self.add_input(name='rho', shape=(nn,), desc='Atmospheric density', units='kg/m**3')
        self.add_input(name='TAS', shape=(nn,), desc='True airspeed', units='m/s')

        # Outputs
        self.add_output(name='q', val=np.zeros(nn), desc='Dynamic pressure', units='Pa')

        # Partials
        ar = np.arange(nn)
        self.declare_partials(of='q', wrt='rho', rows=ar, cols=ar)
        self.declare_partials(of='q', wrt='TAS', rows=ar, cols=ar)

    def compute(self, inputs, outputs):
        outputs['q'] = 0.5 * inputs['rho'] * inputs['TAS'] ** 2

    def compute_partials(self, inputs, partials):
        partials['q', 'rho'] = 0.5 * inputs['TAS'] ** 2
        partials['q', 'TAS'] = inputs['rho'] * inputs['TAS']



import numpy as np

import openmdao.api as om


class CD0Comp(om.ExplicitComponent):
    """ Computes the zero-lift drag coefficient
    """
    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        # Inputs
        self.add_input('mach', shape=(nn,), desc='Mach number', units=None)

        # Outputs
        self.add_output(name='CD0', val=np.zeros(nn), desc='zero-lift drag coefficient', units=None)

        # Jacobian
        ar = np.arange(nn)
        self.declare_partials(of='CD0', wrt='mach', rows=ar, cols=ar)

    def compute(self, inputs, outputs):
        M = inputs['mach']

        idx_low = np.where(M < 1.15)[0]
        idx_high = np.where(M >= 1.15)[0]

        outputs['CD0'][idx_low] = 0.013 + 0.0144 * (1.0 + np.tanh((M[idx_low] - 0.98) / 0.06))
        outputs['CD0'][idx_high] = 0.013 + \
            0.0144 * (1.0 + np.tanh(0.17 / 0.06)) - 0.011 * (M[idx_high] - 1.15)

    def compute_partials(self, inputs, partials):
        M = inputs['mach']

        idx_low = np.where(M < 1.15)[0]
        idx_high = np.where(M >= 1.15)[0]

        k = 50.0 / 3.0

        partials['CD0', 'mach'][idx_low] = 0.24 / (np.cosh(k * (M[idx_low] - 0.98))**2)
        partials['CD0', 'mach'][idx_high] = -0.011



import unittest
import warnings

import numpy as np

from scipy.interpolate import interp1d

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse

import dymos as dm
from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE
from dymos.examples.min_time_climb.min_time_climb_ode import MinTimeClimbODE

from dymos.models.atmosphere import USatm1976Comp
from dymos.examples.min_time_climb.aero import AeroGroup
from dymos.examples.min_time_climb.prop import PropGroup
from dymos.models.eom import FlightPathEOM2D


@use_tempdirs
class TestTimeseriesOutput(unittest.TestCase):

    def test_timeseries_gl(self, test_smaller_timeseries=False):
        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=dm.GaussLobatto(num_segments=8, order=3, compressed=True))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=True)

        phase.add_state('y', fix_initial=True, fix_final=True)

        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', continuity=True, rate_continuity=True, opt=True,
                          units='deg', lower=0.01, upper=179.9, ref=1, ref0=0)

        if test_smaller_timeseries:
            phase.add_parameter('g', opt=True, units='m/s**2', val=9.80665)
        else:
            phase.add_parameter('g', opt=True, units='m/s**2', val=9.80665, include_timeseries=True)

        # Minimize time at the end of the phase
        phase.add_objective('time_phase', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True)

        phase.set_time_val(initial=0, duration=2.0)
        phase.set_state_val('x', (0, 10))
        phase.set_state_val('y', (10, 5))
        phase.set_state_val('v', (0, 9.9))
        phase.set_control_val('theta', (5, 100))
        phase.set_parameter_val('g', 9.80665)

        p.run_driver()

        gd = phase.options['transcription'].grid_data
        state_input_idxs = gd.subset_node_indices['state_input']
        control_input_idxs = gd.subset_node_indices['control_input']
        col_idxs = gd.subset_node_indices['col']

        assert_near_equal(p.get_val('phase0.t'),
                          p.get_val('phase0.timeseries.time')[:, 0])

        assert_near_equal(p.get_val('phase0.t_phase'),
                          p.get_val('phase0.timeseries.time_phase')[:, 0])

        for state in ('x', 'y', 'v'):
            assert_near_equal(p.get_val(f'phase0.states:{state}'),
                              p.get_val(f'phase0.timeseries.{state}')[state_input_idxs])

            assert_near_equal(p.get_val(f'phase0.state_interp.state_col:{state}'),
                              p.get_val(f'phase0.timeseries.{state}')[col_idxs])

        for control in ('theta',):
            assert_near_equal(p.get_val(f'phase0.controls:{control}'),
                              p.get_val(f'phase0.timeseries.{control}')[control_input_idxs])

        for dp in ('g',):
            for i in range(gd.subset_num_nodes['all']):
                if test_smaller_timeseries:
                    with self.assertRaises(KeyError):
                        p.get_val(f'phase0.timeseries.{dp}')
                else:
                    assert_near_equal(p.get_val(f'phase0.parameter_vals:{dp}')[0],
                                      p.get_val(f'phase0.timeseries.{dp}')[i])

        # test simulation
        exp_out = phase.simulate()
        if test_smaller_timeseries:
            with self.assertRaises(KeyError):
                exp_out.get_val(f'phase0.timeseries.{dp}')
        else:  # no error accessing timseries.parameter
            exp_out.get_val(f'phase0.timeseries.{dp}')

    def test_timeseries_gl_smaller_timeseries(self):
        self.test_timeseries_gl(test_smaller_timeseries=True)

    def test_timeseries_radau(self, test_smaller_timeseries=False):

        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=dm.Radau(num_segments=8, order=3, compressed=True))

        phase.timeseries_options['include_state_rates'] = True

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=True)

        phase.add_state('y', fix_initial=True, fix_final=True)

        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', continuity=True, rate_continuity=True, opt=True,
                          units='deg', lower=0.01, upper=179.9, ref=1, ref0=0)

        if test_smaller_timeseries:
            phase.add_parameter('g', opt=True, units='m/s**2', val=9.80665)
        else:
            phase.add_parameter('g', opt=True, units='m/s**2', val=9.80665, include_timeseries=True)

        # Minimize time at the end of the phase
        phase.add_objective('time_phase', loc='final', scaler=10)

        phase.timeseries_options['include_state_rates'] = True

        p.model.options['assembled_jac_type'] = 'csc'
        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True)

        phase.set_time_val(initial=0, duration=2.0)
        phase.set_state_val('x', (0, 10))
        phase.set_state_val('y', (10, 5))
        phase.set_state_val('v', (0, 9.9))
        phase.set_control_val('theta', (5, 100))
        phase.set_parameter_val('g', 9.80665)

        p.run_driver()

        gd = phase.options['transcription'].grid_data
        state_input_idxs = gd.subset_node_indices['state_input']
        control_input_idxs = gd.subset_node_indices['control_input']

        assert_near_equal(p.get_val('phase0.t'),
                          p.get_val('phase0.timeseries.time')[:, 0])

        assert_near_equal(p.get_val('phase0.t_phase'),
                          p.get_val('phase0.timeseries.time_phase')[:, 0])

        for state in ('x', 'y', 'v'):
            assert_near_equal(p.get_val(f'phase0.states:{state}'),
                              p.get_val(f'phase0.timeseries.{state}')[state_input_idxs])

        for control in ('theta',):
            assert_near_equal(p.get_val(f'phase0.controls:{control}'),
                              p.get_val(f'phase0.timeseries.{control}')[control_input_idxs])

        for dp in ('g',):
            for i in range(gd.subset_num_nodes['all']):
                if test_smaller_timeseries:
                    with self.assertRaises(KeyError):
                        p.get_val(f'phase0.timeseries.{dp}')
                else:
                    assert_near_equal(p.get_val(f'phase0.parameters:{dp}')[0],
                                      p.get_val(f'phase0.timeseries.{dp}')[i])

        # test simulation
        exp_out = phase.simulate()
        if test_smaller_timeseries:
            with self.assertRaises(KeyError):
                exp_out.get_val(f'phase0.timeseries.{dp}')
        else:  # no error accessing timseries.parameter
            exp_out.get_val(f'phase0.timeseries.{dp}')

        # Test that the state rates are output in both the radau and solveivp timeseries outputs
        t_sol = p.get_val('phase0.timeseries.time')
        t_sim = exp_out.get_val('phase0.timeseries.time')

        p.model.list_outputs()

        for state_name, rate_name in (('x', 'xdot'), ('y', 'ydot'), ('v', 'vdot')):
            rate_sol = p.get_val(f'phase0.timeseries.{rate_name}')
            rate_sim = exp_out.get_val(f'phase0.timeseries.{rate_name}')
            rate_t_sim = interp1d(t_sim.ravel(), rate_sim.ravel())
            assert_near_equal(rate_t_sim(t_sol), rate_sol, tolerance=1.0E-3)

    def test_timeseries_radau_smaller_timeseries(self):
        self.test_timeseries_radau(test_smaller_timeseries=True)

    def test_timeseries_explicit_shooting(self, test_smaller_timeseries=False):
        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        tx = dm.ExplicitShooting(grid=dm.GaussLobattoGrid(num_segments=3, nodes_per_seg=5))

        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)

        phase.set_time_options(units='s', fix_initial=True, duration_bounds=(1.0, 10.0))

        # automatically discover states
        phase.set_state_options('x', fix_initial=True)
        phase.set_state_options('y', fix_initial=True)
        phase.set_state_options('v', fix_initial=True)

        phase.add_parameter('g', val=1.0, units='m/s**2', opt=True, lower=1, upper=9.80665)
        phase.add_control('theta', val=45.0, units='deg', opt=True, lower=1.0E-6, upper=179.9,
                          ref=90., rate2_continuity=True)

        phase.add_boundary_constraint('x', loc='final', equals=10.0)
        phase.add_boundary_constraint('y', loc='final', equals=5.0)

        p.model.add_subsystem('phase0', phase)

        phase.add_objective('time', loc='final')

        p.setup(force_alloc_complex=True)

        phase.set_time_val(initial=0, duration=2.0)
        phase.set_state_val('x', (0, 10))
        phase.set_state_val('y', (10, 5))
        phase.set_state_val('v', (0, 9.9))
        phase.set_control_val('theta', (5, 100))
        phase.set_parameter_val('g', 9.80665)

        p.run_driver()

        assert_near_equal(np.atleast_2d(p.get_val('phase0.t')).T,
                          p.get_val('phase0.timeseries.time'))

        for state in ('x', 'y', 'v'):
            assert_near_equal(p.get_val(f'phase0.integrator.states_out:{state}'),
                              p.get_val(f'phase0.timeseries.{state}'))

        for control in ('theta',):
            assert_near_equal(p.get_val(f'phase0.control_values:{control}'),
                              p.get_val(f'phase0.timeseries.{control}'))


class MinTimeClimbODEDuplicateOutput(om.Group):

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        self.add_subsystem(name='atmos',
                           subsys=USatm1976Comp(num_nodes=nn),
                           promotes_inputs=['h'])

        self.add_subsystem(name='aero',
                           subsys=AeroGroup(num_nodes=nn),
                           promotes_inputs=['v', 'alpha', 'S'])

        self.connect('atmos.sos', 'aero.sos')
        self.connect('atmos.rho', 'aero.rho')

        self.add_subsystem(name='prop',
                           subsys=PropGroup(num_nodes=nn),
                           promotes_inputs=['h', 'Isp', 'throttle'])

        self.connect('aero.mach', 'prop.mach')

        self.add_subsystem(name='flight_dynamics',
                           subsys=FlightPathEOM2D(num_nodes=nn),
                           promotes_inputs=['m', 'v', 'gam', 'alpha'])

        foo = self.add_subsystem('foo', om.IndepVarComp())
        foo.add_output('rho', val=100 * np.ones(nn), units='g/cm**3')

        self.connect('aero.f_drag', 'flight_dynamics.D')
        self.connect('aero.f_lift', 'flight_dynamics.L')
        self.connect('prop.thrust', 'flight_dynamics.T')


@require_pyoptsparse(optimizer='SLSQP')
def min_time_climb(num_seg=3, transcription_class=dm.Radau, transcription_order=3,
                   force_alloc_complex=False, timeseries_expr=False):

    p = om.Problem(model=om.Group())

    p.driver = om.pyOptSparseDriver()

    tx = transcription_class(num_segments=num_seg, order=transcription_order)

    traj = dm.Trajectory()

    ode_class = MinTimeClimbODEDuplicateOutput if not timeseries_expr else MinTimeClimbODE

    phase = dm.Phase(ode_class=ode_class, transcription=tx)
    traj.add_phase('phase0', phase)

    p.model.add_subsystem('traj', traj)

    phase.set_time_options(fix_initial=True, duration_bounds=(50, 400),
                           duration_ref=100.0)

    phase.add_state('r', fix_initial=True, lower=0, upper=1.0E6,
                    ref=1.0E3, defect_ref=1.0E3, units='m',
                    rate_source='flight_dynamics.r_dot')

    phase.add_state('h', fix_initial=True, lower=0, upper=20000.0,
                    ref=1.0E2, defect_ref=1.0E2, units='m',
                    rate_source='flight_dynamics.h_dot', targets=['h'])

    phase.add_state('v', fix_initial=True, lower=10.0,
                    ref=1.0E2, defect_ref=1.0E2, units='m/s',
                    rate_source='flight_dynamics.v_dot', targets=['v'])

    phase.add_state('gam', fix_initial=True, lower=-1.5, upper=1.5,
                    ref=1.0, defect_ref=1.0, units='rad',
                    rate_source='flight_dynamics.gam_dot', targets=['gam'])

    phase.add_state('m', fix_initial=True, lower=10.0, upper=1.0E5,
                    ref=1.0E3, defect_ref=1.0E3, units='kg',
                    rate_source='prop.m_dot', targets=['m'])

    phase.add_control('alpha', units='deg', lower=-8.0, upper=8.0, scaler=1.0,
                      rate_continuity=True, rate_continuity_scaler=100.0,
                      rate2_continuity=False, targets=['alpha'])

    phase.add_parameter('S', val=49.2386, units='m**2', opt=False, targets=['S'])
    phase.add_parameter('Isp', val=1600.0, units='s', opt=False, targets=['Isp'])
    phase.add_parameter('throttle', val=1.0, opt=False, targets=['throttle'])

    phase.add_boundary_constraint('h', loc='final', equals=20000, scaler=1.0E-3)
    phase.add_boundary_constraint('aero.mach', loc='final', equals=1.0)
    phase.add_boundary_constraint('gam', loc='final', equals=0.0)

    phase.add_path_constraint(name='h', lower=100.0, upper=20000, ref=20000)
    phase.add_path_constraint(name='aero.mach', lower=0.1, upper=1.8)

    # Unnecessary but included to test capability
    phase.add_path_constraint(name='alpha', lower=-8, upper=8)
    phase.add_path_constraint(name='time', lower=0, upper=400)
    phase.add_path_constraint(name='time_phase', lower=0, upper=400)

    # Minimize time at the end of the phase
    phase.add_objective('time', loc='final', ref=1.0)

    # Add all ODE outputs to the timeseries
    phase.add_timeseries_output('*')
    if timeseries_expr:
        phase.add_timeseries_output('R = atmos.pres / (atmos.temp * atmos.rho)')

    p.model.linear_solver = om.DirectSolver()

    p.setup(check=True, force_alloc_complex=force_alloc_complex)

    phase.set_time_val(initial=0, duration=300)
    phase.set_state_val('r', [0.0, 111319.54])
    phase.set_state_val('h', [100.0, 20000.0])
    phase.set_state_val('v', [135.964, 283.159])
    phase.set_state_val('gam', [0.0, 0.0])
    phase.set_state_val('m', [19030.468, 16841.431])
    phase.set_control_val('alpha', [0.0, 0.0])

    return p


@use_tempdirs
class TestDuplicateTimeseriesGlobName(unittest.TestCase):

    def test_duplicate_timeseries_glob_name_radau(self):
        """
        Test that the user gets a warning about multiple timeseries with the same name.
        """
        msg = "Error during configure_timeseries_output_introspection in phase traj.phases.phase0."
        with self.assertRaises(RuntimeError) as e:
            min_time_climb(num_seg=12, transcription_class=dm.Radau, transcription_order=3)
        self.assertEqual(str(e.exception), msg)

    def test_duplicate_timeseries_glob_name_gl(self):
        """
        Test that the user gets a warning about multiple timeseries with the same name.
        """
        msg = "Error during configure_timeseries_output_introspection in phase traj.phases.phase0."
        with self.assertRaises(RuntimeError) as e:
            min_time_climb(num_seg=12, transcription_class=dm.GaussLobatto, transcription_order=3)
        self.assertEqual(str(e.exception), msg)


@use_tempdirs
class TestTimeseriesExprBrachistochrone(unittest.TestCase):

    @staticmethod
    def make_problem_brachistochrone(transcription, polynomial_control=False):
        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=transcription)

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.set_state_options('x', fix_initial=True)
        phase.set_state_options('y', fix_initial=True)
        phase.set_state_options('v', fix_initial=True)

        if polynomial_control:
            phase.add_control('theta', order=1, units='deg', lower=0.01, upper=179.9, control_type='polynomial')
            control_name = 'controls:theta'
        else:
            phase.add_control('theta', continuity=True, rate_continuity=True, opt=True,
                              units='deg', lower=0.01, upper=179.9, ref=1, ref0=0)
            control_name = 'controls:theta'
        phase.add_boundary_constraint('x', loc='final', equals=10.0)
        phase.add_boundary_constraint('y', loc='final', equals=5.0)

        phase.add_parameter('g', opt=False, units='m/s**2', val=9.80665, include_timeseries=True)

        phase.add_objective('time_phase', loc='final', scaler=10)
        phase.add_timeseries_output('z=x*y + x**2', units='m**2')
        phase.add_timeseries_output('f=3*g*cos(theta)**2', units='deg**2')

        p.model.options['assembled_jac_type'] = 'csc'
        p.model.linear_solver = om.DirectSolver()
        p.setup()

        phase.set_time_val(initial=0.0, duration=2.0)
        phase.set_state_val('x', [0, 10])
        phase.set_state_val('y', [10, 5])
        phase.set_state_val('v', [0, 9.9])
        phase.set_control_val('theta', [5, 100])
        phase.set_parameter_val('g', 9.80665)
        return p

    def test_invalid_expr_var(self):
        p = om.Problem(model=om.Group())

        tx = dm.Radau(num_segments=5, order=3, compressed=True)

        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.set_state_options('x', fix_initial=True)
        phase.set_state_options('y', fix_initial=True)
        phase.set_state_options('v', fix_initial=True)

        phase.add_control('theta', continuity=True, rate_continuity=True, opt=True,
                          units='deg', lower=0.01, upper=179.9, ref=1, ref0=0)

        phase.add_parameter('g', opt=False, units='m/s**2', val=9.80665, include_timeseries=True)

        phase.add_timeseries_output('k=units', units='m**2')

        with self.assertRaises(NameError) as e:
            p.setup(check=True)

        expected = ("'phase0.rhs_all.exec_comp' <class ExecComp>: cannot use variable "
                    "name 'units' because it's a reserved keyword.")
        self.assertEqual(str(e.exception), expected)

    def test_input_units(self):
        p = om.Problem(model=om.Group())

        tx = dm.Radau(num_segments=5, order=3, compressed=True)

        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.set_state_options('x', fix_initial=True)
        phase.set_state_options('y', fix_initial=True)
        phase.set_state_options('v', fix_initial=True)

        phase.add_control('theta', continuity=True, rate_continuity=True, opt=True,
                          units='deg', lower=0.01, upper=179.9, ref=1, ref0=0)

        phase.add_parameter('g', opt=False, units='m/s**2', val=9.80665, include_timeseries=True)

        phase.add_timeseries_output('sin_theta=sin(theta)', units='unitless', theta={'units': 'rad'})

        p.setup(check=True)

        phase.set_time_val(initial=0, duration=2.0)
        phase.set_state_val('x', (0, 10))
        phase.set_state_val('y', (10, 5))
        phase.set_state_val('v', (0, 9.9))
        phase.set_control_val('theta', (5, 100))
        phase.set_parameter_val('g', 9.80665)

        p.run_model()

        sin_theta = p.get_val('phase0.timeseries.sin_theta')
        theta = p.get_val('phase0.timeseries.theta', units='rad')

        assert_near_equal(np.sin(theta), sin_theta)

    def test_output_units(self):
        p = om.Problem(model=om.Group())

        tx = dm.Radau(num_segments=5, order=3, compressed=True)

        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.set_state_options('x', fix_initial=True)
        phase.set_state_options('y', fix_initial=True)
        phase.set_state_options('v', fix_initial=True)

        phase.add_control('theta', continuity=True, rate_continuity=True, opt=True,
                          units='deg', lower=0.01, upper=179.9, ref=1, ref0=0)

        phase.add_parameter('g', opt=False, units='m/s**2', val=9.80665, include_timeseries=True)

        phase.add_timeseries_output('sin_theta=sin(theta)', shape=(1,), units='unitless', theta={'units': 'rad', 'shape': (1,)})
        phase.add_timeseries_output('sin_theta2=sin(theta)', sin_theta2={'units': 'unitless'})
        phase.add_timeseries_output('sin_theta3=sin(theta)', shape=(1,), sin_theta3={'shape': (1,)})

        p.setup()

        phase.set_time_val(initial=0.0, duration=2.0)
        phase.set_state_val('x', [0, 10])
        phase.set_state_val('y', [10, 5])
        phase.set_state_val('v', [0, 9.9])
        phase.set_control_val('theta', [5, 100])
        phase.set_parameter_val('g', 9.80665)

        p.run_model()

        var_data = p.model.list_vars(units=True, shape=True, out_stream=None)
        var_data = {meta['prom_name']: meta for abs_path, meta in var_data}

        self.assertEqual('unitless', var_data['phase0.timeseries.sin_theta']['units'])
        self.assertEqual('unitless', var_data['phase0.timeseries.sin_theta']['units'])

    def test_timeseries_expr_radau(self):
        tx = dm.Radau(num_segments=5, order=3, compressed=True)
        p = self.make_problem_brachistochrone(transcription=tx)
        p.run_driver()

        x = p.get_val('phase0.timeseries.x')
        y = p.get_val('phase0.timeseries.y')
        theta = p.get_val('phase0.timeseries.theta')
        g = p.get_val('phase0.timeseries.g')

        z_computed = x * y + x**2
        f_computed = 3 * g * np.cos(theta)**2
        z_ts = p.get_val('phase0.timeseries.z')
        f_ts = p.get_val('phase0.timeseries.f')
        assert_near_equal(z_computed, z_ts, tolerance=1e-12)
        assert_near_equal(f_computed, f_ts, tolerance=1e-12)

    def test_timeseries_expr_radau_polynomial_control(self):
        tx = dm.Radau(num_segments=5, order=3, compressed=True)
        p = self.make_problem_brachistochrone(transcription=tx, polynomial_control=True)
        p.run_driver()

        x = p.get_val('phase0.timeseries.x')
        y = p.get_val('phase0.timeseries.y')
        theta = p.get_val('phase0.timeseries.theta')
        g = p.get_val('phase0.timeseries.g')

        z_computed = x * y + x**2
        f_computed = 3 * g * np.cos(theta)**2
        z_ts = p.get_val('phase0.timeseries.z')
        f_ts = p.get_val('phase0.timeseries.f')
        assert_near_equal(z_computed, z_ts, tolerance=1e-12)
        assert_near_equal(f_computed, f_ts, tolerance=1e-12)

    def test_timeseries_expr_gl(self):
        tx = dm.GaussLobatto(num_segments=5, order=3, compressed=True)
        p = self.make_problem_brachistochrone(transcription=tx)
        p.run_driver()
        x = p.get_val('phase0.timeseries.x')
        y = p.get_val('phase0.timeseries.y')
        theta = p.get_val('phase0.timeseries.theta')
        g = p.get_val('phase0.timeseries.g')

        z_computed = x * y + x**2
        f_computed = 3 * g * np.cos(theta) ** 2

        z_ts = p.get_val('phase0.timeseries.z')
        f_ts = p.get_val('phase0.timeseries.f')
        assert_near_equal(z_computed, z_ts, tolerance=1e-12)
        assert_near_equal(f_computed, f_ts, tolerance=1e-12)

    def test_timeseries_expr_birkhoff(self):
        tx = dm.Birkhoff(num_nodes=21)
        p = self.make_problem_brachistochrone(transcription=tx)
        p.run_driver()
        x = p.get_val('phase0.timeseries.x')
        y = p.get_val('phase0.timeseries.y')
        theta = p.get_val('phase0.timeseries.theta')
        g = p.get_val('phase0.timeseries.g')

        z_computed = x * y + x**2
        f_computed = 3 * g * np.cos(theta) ** 2

        z_ts = p.get_val('phase0.timeseries.z')
        f_ts = p.get_val('phase0.timeseries.f')
        assert_near_equal(z_computed, z_ts, tolerance=1e-12)
        assert_near_equal(f_computed, f_ts, tolerance=1e-12)

    def test_timeseries_expr_gl_polynomial_control(self):
        tx = dm.GaussLobatto(num_segments=5, order=3, compressed=True)
        p = self.make_problem_brachistochrone(transcription=tx, polynomial_control=True)
        p.run_driver()

        x = p.get_val('phase0.timeseries.x')
        y = p.get_val('phase0.timeseries.y')
        theta = p.get_val('phase0.timeseries.theta')
        g = p.get_val('phase0.timeseries.g')

        z_computed = x * y + x**2
        f_computed = 3 * g * np.cos(theta)**2
        z_ts = p.get_val('phase0.timeseries.z')
        f_ts = p.get_val('phase0.timeseries.f')
        assert_near_equal(z_computed, z_ts, tolerance=1e-12)
        assert_near_equal(f_computed, f_ts, tolerance=1e-12)

    def test_timeseries_expr_explicit_shooting(self):

        import warnings

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=om.UnitsWarning)
            tx = dm.ExplicitShooting(grid=dm.GaussLobattoGrid(num_segments=3, nodes_per_seg=5, compressed=True))

            p = self.make_problem_brachistochrone(transcription=tx)
            p.run_driver()
            x = p.get_val('phase0.timeseries.x')
            y = p.get_val('phase0.timeseries.y')
            theta = p.get_val('phase0.timeseries.theta')
            g = p.get_val('phase0.timeseries.g')

            z_computed = x * y + x**2
            f_computed = 3 * g * np.cos(theta) ** 2

            z_ts = p.get_val('phase0.timeseries.z')
            f_ts = p.get_val('phase0.timeseries.f')
            assert_near_equal(z_computed, z_ts, tolerance=1e-12)
            assert_near_equal(f_computed, f_ts, tolerance=1e-12)

    def test_timeseries_expr_explicit_shooting_polynomial_controls(self):
        tx = dm.ExplicitShooting(grid=dm.GaussLobattoGrid(num_segments=3, nodes_per_seg=5, compressed=True))

        p = self.make_problem_brachistochrone(transcription=tx, polynomial_control=True)
        p.run_driver()
        x = p.get_val('phase0.timeseries.x')
        y = p.get_val('phase0.timeseries.y')
        theta = p.get_val('phase0.timeseries.theta')
        g = p.get_val('phase0.timeseries.g')

        z_computed = x * y + x**2
        f_computed = 3 * g * np.cos(theta) ** 2

        z_ts = p.get_val('phase0.timeseries.z')
        f_ts = p.get_val('phase0.timeseries.f')
        assert_near_equal(z_computed, z_ts, tolerance=1e-12)
        assert_near_equal(f_computed, f_ts, tolerance=1e-12)

    def test_timeseries_expr_solve_ivp(self):
        tx = dm.Radau(num_segments=5, order=3, compressed=True)

        p = self.make_problem_brachistochrone(transcription=tx)
        p.run_driver()

        z_ts = p.get_val('phase0.timeseries.z')
        f_ts = p.get_val('phase0.timeseries.f')

        phase0 = p.model._get_subsystem('phase0')
        sim = phase0.simulate(times_per_seg=30)

        z_sim = sim.get_val('phase0.timeseries.z')
        f_sim = sim.get_val('phase0.timeseries.f')
        assert_near_equal(z_ts[-1], z_sim[-1], tolerance=1e-3)
        assert_near_equal(f_ts[-1], f_sim[-1], tolerance=1e-3)

    def test_timeseries_expr_solve_ivp_polynomial_controls(self):
        tx = dm.Radau(num_segments=5, order=3, compressed=True)

        p = self.make_problem_brachistochrone(transcription=tx, polynomial_control=True)
        p.run_driver()

        z_ts = p.get_val('phase0.timeseries.z')
        f_ts = p.get_val('phase0.timeseries.f')

        phase0 = p.model._get_subsystem('phase0')
        sim = phase0.simulate(times_per_seg=30)

        z_sim = sim.get_val('phase0.timeseries.z')
        f_sim = sim.get_val('phase0.timeseries.f')
        assert_near_equal(z_ts[-1], z_sim[-1], tolerance=1e-3)
        assert_near_equal(f_ts[-1], f_sim[-1], tolerance=1e-3)


@use_tempdirs
class TestTimeseriesExprMinTimeClimb(unittest.TestCase):

    def test_timeseries_expr_radau(self):
        p = min_time_climb(transcription_class=dm.Radau, num_seg=12, transcription_order=3, timeseries_expr=True)
        p.run_model()

        pres = p.get_val('traj.phase0.timeseries.pres')
        temp = p.get_val('traj.phase0.timeseries.temp')
        rho = p.get_val('traj.phase0.timeseries.rho')

        R_computed = pres / (temp * rho)
        R_ts = p.get_val('traj.phase0.timeseries.R')
        assert_near_equal(R_computed, R_ts, tolerance=1e-12)

    def test_timeseries_expr_gl(self):
        p = min_time_climb(transcription_class=dm.GaussLobatto, num_seg=12,
                           transcription_order=3, timeseries_expr=True)
        p.run_driver()

        pres = p.get_val('traj.phase0.timeseries.pres')
        temp = p.get_val('traj.phase0.timeseries.temp')
        rho = p.get_val('traj.phase0.timeseries.rho')

        R_computed = pres / (temp * rho)
        R_ts = p.get_val('traj.phase0.timeseries.R')
        assert_near_equal(R_computed, R_ts, tolerance=1e-12)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import unittest
import openmdao.api as om
import dymos as dm
from openmdao.utils.testing_utils import set_env_vars


class TestCheckPartials(unittest.TestCase):

    def brach_explicit_partials(self):
        from dymos.examples.brachistochrone import BrachistochroneODE
        prob = om.Problem()

        prob.driver = om.ScipyOptimizeDriver()

        tx = dm.ExplicitShooting(grid=dm.GaussLobattoGrid(num_segments=3, nodes_per_seg=6, compressed=False))

        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)

        phase.set_time_options(units='s', fix_initial=True,
                               duration_bounds=(1.0, 10.0))

        # automatically discover states
        phase.set_state_options('x', fix_initial=True)
        phase.set_state_options('y', fix_initial=True)
        phase.set_state_options('v', fix_initial=True)

        phase.add_parameter('g', val=1.0, units='m/s**2', opt=True, lower=1,
                            upper=9.80665)
        phase.add_control('theta', val=45.0, units='deg', opt=True, lower=1.0E-6,
                          upper=179.9,
                          ref=90., rate2_continuity=True)

        phase.add_boundary_constraint('x', loc='final', equals=10.0)
        phase.add_boundary_constraint('y', loc='final', equals=5.0)

        prob.model.add_subsystem('phase0', phase)

        phase.add_objective('time', loc='final')

        prob.setup(force_alloc_complex=True)

        prob.set_val('phase0.t_initial', 0.0)
        prob.set_val('phase0.t_duration', 2)
        prob.set_val('phase0.initial_states:x', 0.0)
        prob.set_val('phase0.initial_states:y', 10.0)
        prob.set_val('phase0.initial_states:v', 1.0E-6)
        prob.set_val('phase0.parameters:g', 1.0, units='m/s**2')
        prob.set_val('phase0.controls:theta', phase.interp('theta', ys=[0.01, 90]),
                     units='deg')

        prob.run_model()

        cpd = prob.check_partials(compact_print=True, method='fd', out_stream=None)
        return cpd

    def balanced_field_partials_radau(self):
        from dymos.examples.balanced_field.balanced_field_ode import BalancedFieldODEComp

        p = om.Problem()

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        # First Phase: Brake release to V1 - both engines operable
        br_to_v1 = dm.Phase(ode_class=BalancedFieldODEComp, transcription=dm.Radau(num_segments=3),
                            ode_init_kwargs={'mode': 'runway'})
        br_to_v1.set_time_options(fix_initial=True, duration_bounds=(1, 1000), duration_ref=10.0)
        br_to_v1.add_state('r', fix_initial=True, lower=0, ref=1000.0, defect_ref=1000.0)
        br_to_v1.add_state('v', fix_initial=True, lower=0, ref=100.0, defect_ref=100.0)
        br_to_v1.add_parameter('alpha', val=0.0, opt=False, units='deg')
        br_to_v1.add_timeseries_output('*')

        # Second Phase: Rejected takeoff at V1 - no engines operable
        rto = dm.Phase(ode_class=BalancedFieldODEComp, transcription=dm.Radau(num_segments=3),
                       ode_init_kwargs={'mode': 'runway'})
        rto.set_time_options(fix_initial=False, duration_bounds=(1, 1000), duration_ref=1.0)
        rto.add_state('r', fix_initial=False, lower=0, ref=1000.0, defect_ref=1000.0)
        rto.add_state('v', fix_initial=False, lower=0, ref=100.0, defect_ref=100.0)
        rto.add_parameter('alpha', val=0.0, opt=False, units='deg')
        rto.add_timeseries_output('*')

        # Third Phase: V1 to Vr - single engine operable
        v1_to_vr = dm.Phase(ode_class=BalancedFieldODEComp, transcription=dm.Radau(num_segments=3),
                            ode_init_kwargs={'mode': 'runway'})
        v1_to_vr.set_time_options(fix_initial=False, duration_bounds=(1, 1000), duration_ref=1.0)
        v1_to_vr.add_state('r', fix_initial=False, lower=0, ref=1000.0, defect_ref=1000.0)
        v1_to_vr.add_state('v', fix_initial=False, lower=0, ref=100.0, defect_ref=100.0)
        v1_to_vr.add_parameter('alpha', val=0.0, opt=False, units='deg')
        v1_to_vr.add_timeseries_output('*')

        # Fourth Phase: Rotate - single engine operable
        rotate = dm.Phase(ode_class=BalancedFieldODEComp, transcription=dm.Radau(num_segments=3),
                          ode_init_kwargs={'mode': 'runway'})
        rotate.set_time_options(fix_initial=False, duration_bounds=(1.0, 5), duration_ref=1.0)
        rotate.add_state('r', fix_initial=False, lower=0, ref=1000.0, defect_ref=1000.0)
        rotate.add_state('v', fix_initial=False, lower=0, ref=100.0, defect_ref=100.0)
        rotate.add_control('alpha', order=1, opt=True, units='deg', lower=0, upper=10,
                           ref=10, val=[0, 10], control_type='polynomial')
        rotate.add_timeseries_output('*')

        # Fifth Phase: Climb to target speed and altitude at end of runway.
        climb = dm.Phase(ode_class=BalancedFieldODEComp, transcription=dm.Radau(num_segments=5),
                         ode_init_kwargs={'mode': 'climb'})
        climb.set_time_options(fix_initial=False, duration_bounds=(1, 100), duration_ref=1.0)
        climb.add_state('r', fix_initial=False, lower=0, ref=1000.0, defect_ref=1000.0)
        climb.add_state('h', fix_initial=True, lower=0, ref=1.0, defect_ref=1.0)
        climb.add_state('v', fix_initial=False, lower=0, ref=100.0, defect_ref=100.0)
        climb.add_state('gam', fix_initial=True, lower=0, ref=0.05, defect_ref=0.05)
        climb.add_control('alpha', opt=True, units='deg', lower=-10, upper=15, ref=10)
        climb.add_timeseries_output('*')

        # Instantiate the trajectory and add phases
        traj = dm.Trajectory()
        p.model.add_subsystem('traj', traj)
        traj.add_phase('br_to_v1', br_to_v1)
        traj.add_phase('rto', rto)
        traj.add_phase('v1_to_vr', v1_to_vr)
        traj.add_phase('rotate', rotate)
        traj.add_phase('climb', climb)

        # Add parameters common to multiple phases to the trajectory
        traj.add_parameter('m', val=174200., opt=False, units='lbm',
                           desc='aircraft mass',
                           targets={'br_to_v1': ['m'], 'v1_to_vr': ['m'], 'rto': ['m'],
                                    'rotate': ['m'], 'climb': ['m']})

        traj.add_parameter('T_nominal', val=27000 * 2, opt=False, units='lbf', static_target=True,
                           desc='nominal aircraft thrust',
                           targets={'br_to_v1': ['T']})

        traj.add_parameter('T_engine_out', val=27000, opt=False, units='lbf', static_target=True,
                           desc='thrust under a single engine',
                           targets={'v1_to_vr': ['T'], 'rotate': ['T'], 'climb': ['T']})

        traj.add_parameter('T_shutdown', val=0.0, opt=False, units='lbf', static_target=True,
                           desc='thrust when engines are shut down for rejected takeoff',
                           targets={'rto': ['T']})

        traj.add_parameter('mu_r_nominal', val=0.03, opt=False, units=None, static_target=True,
                           desc='nominal runway friction coefficient',
                           targets={'br_to_v1': ['mu_r'], 'v1_to_vr': ['mu_r'], 'rotate': ['mu_r']})

        traj.add_parameter('mu_r_braking', val=0.3, opt=False, units=None, static_target=True,
                           desc='runway friction coefficient under braking',
                           targets={'rto': ['mu_r']})

        traj.add_parameter('h_runway', val=0., opt=False, units='ft',
                           desc='runway altitude',
                           targets={'br_to_v1': ['h'], 'v1_to_vr': ['h'], 'rto': ['h'],
                                    'rotate': ['h']})

        traj.add_parameter('rho', val=1.225, opt=False, units='kg/m**3', static_target=True,
                           desc='atmospheric density',
                           targets={'br_to_v1': ['rho'], 'v1_to_vr': ['rho'], 'rto': ['rho'],
                                    'rotate': ['rho']})

        traj.add_parameter('S', val=124.7, opt=False, units='m**2', static_target=True,
                           desc='aerodynamic reference area',
                           targets={'br_to_v1': ['S'], 'v1_to_vr': ['S'], 'rto': ['S'],
                                    'rotate': ['S'], 'climb': ['S']})

        traj.add_parameter('CD0', val=0.03, opt=False, units=None, static_target=True,
                           desc='zero-lift drag coefficient',
                           targets={f'{phase}': ['CD0'] for phase in ['br_to_v1', 'v1_to_vr',
                                                                      'rto', 'rotate', 'climb']})

        traj.add_parameter('AR', val=9.45, opt=False, units=None, static_target=True,
                           desc='wing aspect ratio',
                           targets={f'{phase}': ['AR'] for phase in ['br_to_v1', 'v1_to_vr',
                                                                     'rto', 'rotate', 'climb']})

        traj.add_parameter('e', val=801, opt=False, units=None, static_target=True,
                           desc='Oswald span efficiency factor',
                           targets={f'{phase}': ['e'] for phase in ['br_to_v1', 'v1_to_vr',
                                                                    'rto', 'rotate', 'climb']})

        traj.add_parameter('span', val=35.7, opt=False, units='m', static_target=True,
                           desc='wingspan',
                           targets={f'{phase}': ['span'] for phase in ['br_to_v1', 'v1_to_vr',
                                                                       'rto', 'rotate', 'climb']})

        traj.add_parameter('h_w', val=1.0, opt=False, units='m', static_target=True,
                           desc='height of wing above CG',
                           targets={f'{phase}': ['h_w'] for phase in ['br_to_v1', 'v1_to_vr',
                                                                      'rto', 'rotate', 'climb']})

        traj.add_parameter('CL0', val=0.5, opt=False, units=None, static_target=True,
                           desc='zero-alpha lift coefficient',
                           targets={f'{phase}': ['CL0'] for phase in ['br_to_v1', 'v1_to_vr',
                                                                      'rto', 'rotate', 'climb']})

        traj.add_parameter('CL_max', val=2.0, opt=False, units=None, static_target=True,
                           desc='maximum lift coefficient for linear fit',
                           targets={f'{phase}': ['CL_max'] for phase in ['br_to_v1', 'v1_to_vr',
                                                                         'rto', 'rotate', 'climb']})

        traj.add_parameter('alpha_max', val=10.0, opt=False, units='deg', static_target=True,
                           desc='angle of attack at maximum lift',
                           targets={f'{phase}': ['alpha_max'] for phase in ['br_to_v1', 'v1_to_vr',
                                                                            'rto', 'rotate', 'climb']})

        # Standard "end of first phase to beginning of second phase" linkages
        # Alpha changes from being a parameter in v1_to_vr to a polynomial control
        # in rotate, to a dynamic control in `climb`.
        traj.link_phases(['br_to_v1', 'v1_to_vr'], vars=['time', 'r', 'v'])
        traj.link_phases(['v1_to_vr', 'rotate'], vars=['time', 'r', 'v', 'alpha'])
        traj.link_phases(['rotate', 'climb'], vars=['time', 'r', 'v', 'alpha'])
        traj.link_phases(['br_to_v1', 'rto'], vars=['time', 'r', 'v'])

        # Less common "final value of r must be the match at ends of two phases".
        traj.add_linkage_constraint(phase_a='rto', var_a='r', loc_a='final',
                                    phase_b='climb', var_b='r', loc_b='final',
                                    ref=1000)

        # Define the constraints and objective for the optimal control problem
        v1_to_vr.add_boundary_constraint('v_over_v_stall', loc='final', lower=1.2, ref=100)

        rto.add_boundary_constraint('v', loc='final', equals=0., ref=100, linear=True)

        rotate.add_boundary_constraint('F_r', loc='final', equals=0, ref=100000)

        climb.add_boundary_constraint('h', loc='final', equals=35, ref=35, units='ft', linear=True)
        climb.add_boundary_constraint('gam', loc='final', equals=5, ref=5, units='deg', linear=True)
        climb.add_path_constraint('gam', lower=0, upper=5, ref=5, units='deg')
        climb.add_boundary_constraint('v_over_v_stall', loc='final', lower=1.25, ref=1.25)

        rto.add_objective('r', loc='final', ref=1000.0)

        #
        # Setup the problem and set the initial guess
        #
        p.setup(check=True)

        p.set_val('traj.br_to_v1.t_initial', 0)
        p.set_val('traj.br_to_v1.t_duration', 35)
        p.set_val('traj.br_to_v1.states:r', br_to_v1.interp('r', [0, 2500.0]))
        p.set_val('traj.br_to_v1.states:v', br_to_v1.interp('v', [0, 100.0]))
        p.set_val('traj.br_to_v1.parameters:alpha', 0, units='deg')

        p.set_val('traj.v1_to_vr.t_initial', 35)
        p.set_val('traj.v1_to_vr.t_duration', 35)
        p.set_val('traj.v1_to_vr.states:r', v1_to_vr.interp('r', [2500, 300.0]))
        p.set_val('traj.v1_to_vr.states:v', v1_to_vr.interp('v', [100, 110.0]))
        p.set_val('traj.v1_to_vr.parameters:alpha', 0.0, units='deg')

        p.set_val('traj.rto.t_initial', 35)
        p.set_val('traj.rto.t_duration', 35)
        p.set_val('traj.rto.states:r', rto.interp('r', [2500, 5000.0]))
        p.set_val('traj.rto.states:v', rto.interp('v', [110, 0]))
        p.set_val('traj.rto.parameters:alpha', 0.0, units='deg')

        p.set_val('traj.rotate.t_initial', 70)
        p.set_val('traj.rotate.t_duration', 5)
        p.set_val('traj.rotate.states:r', rotate.interp('r', [1750, 1800.0]))
        p.set_val('traj.rotate.states:v', rotate.interp('v', [80, 85.0]))
        p.set_val('traj.rotate.controls:alpha', 0.0, units='deg')

        p.set_val('traj.climb.t_initial', 75)
        p.set_val('traj.climb.t_duration', 15)
        p.set_val('traj.climb.states:r', climb.interp('r', [5000, 5500.0]), units='ft')
        p.set_val('traj.climb.states:v', climb.interp('v', [160, 170.0]), units='kn')
        p.set_val('traj.climb.states:h', climb.interp('h', [0, 35.0]), units='ft')
        p.set_val('traj.climb.states:gam', climb.interp('gam', [0, 5.0]), units='deg')
        p.set_val('traj.climb.controls:alpha', 5.0, units='deg')

        p.run_model()

        cpd = p.check_partials(compact_print=True, method='fd', out_stream=None)

        # Filter out those that include rhs_all in the path
        cpd = {path: data for path, data in cpd.items() if 'rhs_all' not in path}

        return cpd

    def min_time_climb_partials_gl(self):
        from dymos.examples.min_time_climb.min_time_climb_ode import MinTimeClimbODE

        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        #
        # Instantiate the trajectory and phase
        #
        traj = dm.Trajectory()

        phase = dm.Phase(ode_class=MinTimeClimbODE,
                         transcription=dm.GaussLobatto(num_segments=15, compressed=False))

        traj.add_phase('phase0', phase)

        p.model.add_subsystem('traj', traj)

        #
        # Set the options on the optimization variables
        # Note the use of explicit state units here since much of the ODE uses imperial units
        # and we prefer to solve this problem using metric units.
        #
        phase.set_time_options(fix_initial=True, duration_bounds=(50, 400),
                               duration_ref=100.0)

        phase.add_state('r', fix_initial=True, lower=0, upper=1.0E6, units='m',
                        ref=1.0E3, defect_ref=1.0E3,
                        rate_source='flight_dynamics.r_dot')

        phase.add_state('h', fix_initial=True, lower=0, upper=20000.0, units='m',
                        ref=1.0E2, defect_ref=1.0E2,
                        rate_source='flight_dynamics.h_dot')

        phase.add_state('v', fix_initial=True, lower=10.0, units='m/s',
                        ref=1.0E2, defect_ref=1.0E2,
                        rate_source='flight_dynamics.v_dot')

        phase.add_state('gam', fix_initial=True, lower=-1.5, upper=1.5, units='rad',
                        ref=1.0, defect_ref=1.0,
                        rate_source='flight_dynamics.gam_dot')

        phase.add_state('m', fix_initial=True, lower=10.0, upper=1.0E5, units='kg',
                        ref=1.0E3, defect_ref=1.0E3,
                        rate_source='prop.m_dot')

        phase.add_control('alpha', units='deg', lower=-8.0, upper=8.0, scaler=1.0,
                          rate_continuity=True, rate_continuity_scaler=100.0,
                          rate2_continuity=False)

        phase.add_parameter('S', val=49.2386, units='m**2', opt=False, targets=['S'])
        phase.add_parameter('Isp', val=1600.0, units='s', opt=False, targets=['Isp'])
        phase.add_parameter('throttle', val=1.0, opt=False, targets=['throttle'])

        #
        # Setup the boundary and path constraints
        #
        phase.add_boundary_constraint('h', loc='final', equals=20000, scaler=1.0E-3)
        phase.add_boundary_constraint('aero.mach', loc='final', equals=1.0)
        phase.add_boundary_constraint('gam', loc='final', equals=0.0)

        phase.add_path_constraint(name='h', lower=100.0, upper=20000, ref=20000)
        phase.add_path_constraint(name='aero.mach', lower=0.1, upper=1.8)

        # Minimize time at the end of the phase
        phase.add_objective('time', loc='final', ref=1.0)

        p.model.linear_solver = om.DirectSolver()

        #
        # Setup the problem and set the initial guess
        #
        p.setup(check=True)

        p['traj.phase0.t_initial'] = 0.0
        p['traj.phase0.t_duration'] = 500

        p.set_val('traj.phase0.states:r', phase.interp('r', [0.0, 50000.0]))
        p.set_val('traj.phase0.states:h', phase.interp('h', [100.0, 20000.0]))
        p.set_val('traj.phase0.states:v', phase.interp('v', [135.964, 283.159]))
        p.set_val('traj.phase0.states:gam', phase.interp('gam', [0.0, 0.0]))
        p.set_val('traj.phase0.states:m', phase.interp('m', [19030.468, 10000.]))
        p.set_val('traj.phase0.controls:alpha', phase.interp('alpha', [0.0, 0.0]))

        p.run_model()

        cpd = p.check_partials(compact_print=True, method='fd', out_stream=None)

        # Filter out those that include rhs_all in the path
        cpd = {path: data for path, data in cpd.items() if 'rhs_disc' not in path and 'rhs_col' not in path}

        return cpd

    def test_check_partials_yes(self):
        """
        Run check_partials on a series of dymos problems and verify that partials information
        is displayed for core Dymos components when DYMOS_CHECK_PARTIALS == 'True'.
        """
        cp_save = dm.options['include_check_partials']
        dm.options['include_check_partials'] = True

        cases = [self.brach_explicit_partials,
                 self.balanced_field_partials_radau,
                 self.min_time_climb_partials_gl]

        partials = {}
        for c in cases:
            partials.update(c())

        dm.options['include_check_partials'] = cp_save

        assert len(partials.keys()) > 0

    @set_env_vars(CI='0')  # Make sure _no_check_partials isn't disabled
    def test_check_partials_no(self):
        """
        Run check_partials on a series of dymos problems and verify that partials information
        is not displayed for core Dymos components when DYMOS_CHECK_PARTIALS == 'False'.
        """
        cp_save = dm.options['include_check_partials']
        dm.options['include_check_partials'] = False

        cases = [self.brach_explicit_partials,
                 self.balanced_field_partials_radau,
                 self.min_time_climb_partials_gl]

        partials = {}
        for c in cases:
            partials.update(c())

        dm.options['include_check_partials'] = cp_save

        # Only `phase.ode` should show up in in the partials keys.
        self.assertSetEqual(set(partials.keys()), {'phase0.ode'})


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import numpy as np

import openmdao.api as om


class SFCComp(om.ExplicitComponent):
    """ Compute the specific fuel consumption based on the altitude
    and the sea-level specific fuel consumption.
    """
    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        self.Ka = 1.5E-4 * 9.80665  # Altitude correction factor

        # Inputs
        self.add_input(name='tsfc_sl',
                       desc='sea-level specific fuel consumption',
                       units='1/s')

        self.add_input(name='alt',
                       shape=(nn,),
                       desc='altitude',
                       units='m')

        # Outputs
        self.add_output(name='tsfc',
                        val=np.zeros(nn),
                        desc='specific fuel consumption',
                        units='1/s')

        # Partials
        ar = np.arange(nn)
        self.declare_partials('tsfc', 'tsfc_sl', rows=ar, cols=np.zeros(nn), val=1.0)
        self.declare_partials('tsfc', 'alt', rows=ar, cols=ar, val=-1.0E-6 * self.Ka)

    def compute(self, inputs, outputs):
        outputs['tsfc'] = inputs['tsfc_sl'] - 1.0E-6 * self.Ka * inputs['alt']



import os

from openmdao.api import OptionsDictionary


options = OptionsDictionary()

_env_check_partials = os.environ.get('DYMOS_CHECK_PARTIALS', '0')
_icp_default = _env_check_partials.lower() in ('1', 'yes', 'true')


def _removed_option(name, value):
    if value is not None:
        raise ValueError(f'Option {name} has been replaced by '
                         'Phase.timeseries_options["use_prefix"].')


options.declare('include_check_partials', default=_icp_default, types=bool,
                desc='If True, include dymos components when checking partials.')

options.declare('plots', default='bokeh', values=['matplotlib', 'bokeh'],
                desc='The plot library used to generate output plots for Dymos.')

options.declare('notebook_mode', default=False, types=bool,
                desc='If True, provide notebook-enhanced plots and outputs.')

options.declare('use_timeseries_prefix', default=None, allow_none=True,
                check_valid=_removed_option,
                desc='Note: This option is no longer valid and has been '
                     'replaced by Phase.timeseries_options["use_prefix"].')



import os
import pathlib
import subprocess
import sys
import unittest

try:
    import pycodestyle
except ImportError:
    pycodestyle = None

import dymos

EXCLUDE_FILES = ['crm_data.py']


def _discover_python_files(path):
    """
    Recursively walk through the path and find all python files.

    Parameters
    ----------
    path : str
        The path to be traversed.

    Returns
    -------
    list
        All the python files contained within the given path

    """
    python_files = []
    for root, dirs, files in os.walk(path):
        python_files += [os.path.join(root, file) for file in files if file.endswith('.py')]
    return python_files


def _get_tracked_python_files(git_root):
    """
    Given a git root directory

    Parameters
    ----------
    git_root : str or Path
        The root directory of the git repository.

    Returns
    -------
    set
        Python files in the given git directory that a tracked by git.
    """
    _git_root = str(git_root)
    file_list = _discover_python_files(_git_root)
    untracked_set = set(
        subprocess.run(
            ['git', 'ls-files', _git_root, '--exclude-standard', '--others'],
            stdout=subprocess.PIPE,
            text=True
        ).stdout.splitlines()
    )
    untracked_set = {str(pathlib.Path(f).absolute()) for f in untracked_set if f.endswith('.py')}
    return set(file_list) - untracked_set


@unittest.skipIf(pycodestyle is None, "This test requires pycodestyle")
class TestPyCodeStyle(unittest.TestCase):

    def test_pycodestyle(self):
        """ Tests that all files in this, directory, the parent directory, and test
        sub-directories are PEP8 compliant.

        Notes
        -----
        max_line_length has been set to 130 for this test.
        """
        dymos_path = os.path.split(dymos.__file__)[0]
        pyfiles = _discover_python_files(dymos_path)

        files_to_check = _get_tracked_python_files(dymos_path)

        style = pycodestyle.StyleGuide(ignore=['E226',  # missing whitespace around arithmetic operator
                                               'E241',  # multiple spaces after ','
                                               'W504',  # line break after binary operator
                                               'W605',  # invalid escape sequence
                                               'E722',  # do not use bare except
                                               'E741'   # ambiguous variable name
                                               ])
        style.options.max_line_length = 130

        # the report writes most failures to stdout which is swallowed by testflo by
        # default, so temporarily replace stdout with a StringIO so we can include the failure
        # descriptions in the failure message.
        from io import StringIO
        save_out = sys.stdout
        save_err = sys.stderr
        try:
            sys.stdout = buff_out = StringIO()
            sys.stdout = buff_err = StringIO()
            report = style.check_files(files_to_check)
        finally:
            sys.stdout = save_out
            sys.stderr = save_err
            fails = buff_out.getvalue()
            fails += '\n' + buff_err.getvalue()

        if report.total_errors > 0:
            self.fail(f"Found {report.total_errors} pycodestyle errors:\n{fails}")


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import unittest

from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse

import numpy as np

import openmdao.api as om
from openmdao.components.interp_util.interp import InterpND
from openmdao.utils.assert_utils import assert_near_equal

import dymos as dm
from dymos.models.atmosphere.atmos_1976 import USatm1976Data


class CannonballSizeComp(om.ExplicitComponent):
    """
    Compute the area and mass of a cannonball with a given radius and density.

    Notes
    -----
    This component is not vectorized with 'num_nodes' as is the usual way
    with Dymos, but is instead intended to compute a scalar mass and reference
    area from scalar radius and density inputs. This component does not reside
    in the ODE but instead its outputs are connected to the trajectory via
    input design parameters.
    """

    def setup(self):
        self.add_input(name='radius', val=1.0, desc='cannonball radius', units='m')
        self.add_input(name='dens', val=7870., desc='cannonball density', units='kg/m**3')

        self.add_output(name='mass', shape=(1,), desc='cannonball mass', units='kg')
        self.add_output(name='S', shape=(1,), desc='aerodynamic reference area', units='m**2')

        self.declare_partials(of='mass', wrt='dens')
        self.declare_partials(of='mass', wrt='radius')

        self.declare_partials(of='S', wrt='radius')

    def compute(self, inputs, outputs):
        radius = inputs['radius']
        dens = inputs['dens']

        outputs['mass'] = (4 / 3.) * dens * np.pi * radius ** 3
        outputs['S'] = np.pi * radius ** 2

    def compute_partials(self, inputs, partials):
        radius = inputs['radius']
        dens = inputs['dens']

        partials['mass', 'dens'] = (4 / 3.) * np.pi * radius ** 3
        partials['mass', 'radius'] = 4. * dens * np.pi * radius ** 2

        partials['S', 'radius'] = 2 * np.pi * radius


class CannonballODE(om.ExplicitComponent):
    """
    Cannonball ODE assuming flat earth and accounting for air resistance
    """

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        # static parameters
        self.add_input('m', units='kg')
        self.add_input('S', units='m**2')
        # 0.5 good assumption for a sphere
        self.add_input('CD', 0.5)

        # time varying inputs
        self.add_input('h', units='m', shape=nn)
        self.add_input('v', units='m/s', shape=nn)
        self.add_input('gam', units='rad', shape=nn)

        # state rates
        self.add_output('v_dot', shape=nn, units='m/s**2', tags=['dymos.state_rate_source:v'])
        self.add_output('gam_dot', shape=nn, units='rad/s', tags=['dymos.state_rate_source:gam'])
        self.add_output('h_dot', shape=nn, units='m/s', tags=['dymos.state_rate_source:h'])
        self.add_output('r_dot', shape=nn, units='m/s', tags=['dymos.state_rate_source:r'])
        self.add_output('ke', shape=nn, units='J')

        # Ask OpenMDAO to compute the partial derivatives using finite-difference
        # with a partial coloring algorithm for improved performance, and use
        # a graph coloring algorithm to automatically detect the sparsity pattern.
        self.declare_coloring(wrt='*', method='fd')

        alt_data = USatm1976Data.alt * om.unit_conversion('ft', 'm')[0]
        rho_data = USatm1976Data.rho * om.unit_conversion('slug/ft**3', 'kg/m**3')[0]
        self.rho_interp = InterpND(points=np.array(alt_data),
                                   values=np.array(rho_data),
                                   method='slinear', extrapolate=True)

    def compute(self, inputs, outputs):

        gam = inputs['gam']
        v = inputs['v']
        h = inputs['h']
        m = inputs['m']
        S = inputs['S']
        CD = inputs['CD']

        GRAVITY = 9.80665  # m/s**2

        rho = self.rho_interp.interpolate(inputs['h'])

        q = 0.5 * rho * inputs['v'] ** 2
        qS = q * S
        D = qS * CD
        cgam = np.cos(gam)
        sgam = np.sin(gam)
        outputs['v_dot'] = - D / m - GRAVITY * sgam
        outputs['gam_dot'] = -(GRAVITY / v) * cgam
        outputs['h_dot'] = v * sgam
        outputs['r_dot'] = v * cgam
        outputs['ke'] = 0.5 * m * v ** 2


@use_tempdirs
class TestTwoPhaseCannonballForDocs(unittest.TestCase):

    @require_pyoptsparse(optimizer='SLSQP')
    def make_problem(self, connected=False):

        #############################################
        # Setup the Dymos problem
        #############################################

        p = om.Problem(model=om.Group())

        p.driver = om.pyOptSparseDriver()
        p.driver.options['optimizer'] = 'SLSQP'
        p.driver.declare_coloring()

        p.model.add_subsystem('size_comp', CannonballSizeComp(),
                              promotes_inputs=['radius', 'dens'])
        p.model.set_input_defaults('dens', val=7.87, units='g/cm**3')
        p.model.add_design_var('radius', lower=0.01, upper=0.10,
                               ref0=0.01, ref=0.10, units='m')

        traj = dm.Trajectory()
        p.model.add_subsystem('traj', traj)

        transcription = dm.Birkhoff(num_nodes=10)
        ascent = dm.Phase(ode_class=CannonballODE, transcription=transcription)

        ascent = traj.add_phase('ascent', ascent)

        # All initial states except flight path angle are fixed
        # Final flight path angle is fixed (we will set it to zero
        # so that the phase ends at apogee).
        # The output of the ODE which provides the rate source for each state
        # is obtained from the tags used on those outputs in the ODE.
        # The units of the states are automatically inferred by multiplying the units
        # of those rates by the time units.
        ascent.set_time_options(fix_initial=True, duration_bounds=(1, 100),
                                duration_ref=100, units='s')
        ascent.set_state_options('r', fix_initial=True, fix_final=False)
        ascent.set_state_options('h', fix_initial=True, fix_final=False)
        ascent.set_state_options('gam', fix_initial=False, fix_final=True)
        ascent.set_state_options('v', fix_initial=False, fix_final=False)

        ascent.add_parameter('S', units='m**2', static_target=True)
        ascent.add_parameter('m', units='kg', static_target=True)

        # Limit the muzzle energy
        ascent.add_boundary_constraint('ke', loc='initial',
                                       upper=400000, lower=0, ref=100000)

        # Second Phase (descent)
        transcription = dm.Birkhoff(num_nodes=10)
        descent = dm.Phase(ode_class=CannonballODE, transcription=transcription)

        traj.add_phase('descent', descent)

        # All initial states and time are free, since
        #    they will be linked to the final states of ascent.
        # Final altitude is fixed, because we will set
        #    it to zero so that the phase ends at ground impact)
        descent.set_time_options(duration_bounds=(.5, 100),
                                 duration_ref=100, units='s')
        descent.add_state('r')
        descent.add_state('h', fix_initial=False, fix_final=True)
        descent.add_state('gam', fix_initial=False, fix_final=False)
        descent.add_state('v', fix_initial=False, fix_final=False)

        descent.add_parameter('S', units='m**2', static_target=True)
        descent.add_parameter('m', units='kg', static_target=True)

        descent.add_objective('r', loc='final', scaler=-1.0)

        # Add internally-managed design parameters to the trajectory.
        traj.add_parameter('CD',
                           targets={'ascent': ['CD'], 'descent': ['CD']},
                           val=0.5, units=None, opt=False, static_target=True)

        # Add externally-provided design parameters to the trajectory.
        # In this case, we connect 'm' to pre-existing input parameters
        # named 'm' in each phase.
        traj.add_parameter('m', units='kg', val=1.0,
                           targets={'ascent': 'm', 'descent': 'm'}, static_target=True)

        # In this case, by omitting targets, we're connecting these
        # parameters to parameters with the same name in each phase.
        traj.add_parameter('S', units='m**2', val=0.005, static_target=True)

        # Link Phases (link time and all state variables)
        traj.link_phases(phases=['ascent', 'descent'], vars=['*'], connected=connected)

        # Issue Connections
        p.model.connect('size_comp.mass', 'traj.parameters:m')
        p.model.connect('size_comp.S', 'traj.parameters:S')

        # A linear solver at the top level can improve performance.
        p.model.linear_solver = om.DirectSolver()

        # Finish Problem Setup
        p.setup()

        #############################################
        # Set constants and initial guesses
        #############################################
        p.set_val('radius', 0.05, units='m')
        p.set_val('dens', 7.87, units='g/cm**3')

        traj.set_parameter_val('CD', 0.5)

        ascent.set_time_val(initial=0.0, duration=10.0)

        ascent.set_state_val('r', [0, 100])
        ascent.set_state_val('h', [0, 100])
        ascent.set_state_val('v', [200, 150])
        ascent.set_state_val('gam', [25, 0], units='deg')

        descent.set_time_val(initial=10.0, duration=10.0)

        descent.set_state_val('r', [100, 200])
        descent.set_state_val('h', [100, 0])
        descent.set_state_val('v', [150, 200])
        descent.set_state_val('gam', [0, -45], units='deg')

        dm.run_problem(p)
        assert_near_equal(p.get_val('traj.descent.states:r')[-1],
                          3183.25, tolerance=1.0E-4)

    def test_two_phase_cannonball_birkhoff_unconnected(self):
        self.make_problem(connected=False)

    def test_two_phase_cannonball_birkhoff_connected(self):
        self.make_problem(connected=True)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



"""Definition of the States passthru component."""
import numpy as np

from openmdao.core.explicitcomponent import ExplicitComponent
from dymos.transcriptions.grid_data import GridData
from dymos._options import options as dymos_options


class StatesComp(ExplicitComponent):
    """
    This component provides inputs for states and outputs their values during Picard iteration.

    This component serves to accept variables 'states:{state_name}' and echo them out as
    'state_val:{state_name}'.  This is necessary when using NonlinearBlockGS to converge
    the states.

    Parameters
    ----------
    **kwargs : dict
        Arguments to be passed to the component initialization method.

    Attributes
    ----------
    _vars : dict
        Container mapping name of variables to be muxed with additional data.
    _input_names : dict
        Container mapping name of variables to be muxed with associated inputs.
    """

    def __init__(self, **kwargs):
        """
        Instantiate MuxComp and populate private members.
        """
        super().__init__(**kwargs)

        self._no_check_partials = not dymos_options['include_check_partials']

    def initialize(self):
        """
        Declare component options.
        """
        self.options.declare(
            'grid_data', types=GridData,
            desc='Container object for grid info')

        self.options.declare(
            'state_options', types=dict,
            desc='Dictionary of state names/options for the phase')

    def configure_io(self, phase):
        """
        I/O creation is delayed until configure so we can determine shape and units.

        Parameters
        ----------
        phase : Phase
            The phase object that contains this collocation comp.
        """
        gd = self.options['grid_data']
        num_nodes = gd.subset_num_nodes['all']
        state_options = self.options['state_options']

        self.var_names = var_names = {}
        for state_name, options in state_options.items():
            var_names[state_name] = {
                'current_state': f'state_val:{state_name}',
                'next_state': f'states:{state_name}'
            }

        ar = np.arange(num_nodes, dtype=int)
        for state_name, options in state_options.items():
            shape = options['shape']
            units = options['units']

            var_names = self.var_names[state_name]

            self.add_input(
                name=var_names['next_state'],
                shape=(num_nodes,) + shape,
                desc=f'Value of state {state_name} available as an input.',
                units=units)

            self.add_output(
                name=var_names['current_state'],
                shape=(num_nodes,) + shape,
                desc=f'Value of the state {state_name} to be used in the ODE evaluation.',
                units=units
            )

            self.declare_partials(of=var_names['current_state'],
                                  wrt=var_names['next_state'],
                                  rows=ar, cols=ar, val=1.0)

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        """
        Mux the inputs into the appropriate outputs.

        Parameters
        ----------
        inputs : Vector
            Unscaled, dimensional input variables read via inputs[key].
        outputs : Vector
            Unscaled, dimensional output variables read via outputs[key].
        discrete_inputs : Vector
            Discrete input variables read via discrete_inputs[key].
        discrete_outputs : Vector
            Discrete output variables read via discrete_outputs[key].
        """
        outputs.set_val(inputs.asarray())



import unittest

import numpy as np

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs

from dymos.utils.testing_utils import assert_check_partials
from dymos.examples.aircraft_steady_flight.propulsion.propulsion_group import PropulsionGroup


@use_tempdirs
class TestPropulsionComp(unittest.TestCase):

    @classmethod
    def setUp(self):
        self.n = 10

        self.p = om.Problem(model=om.Group())

        ivc = self.p.model.add_subsystem('ivc', om.IndepVarComp(), promotes_outputs=['*'])

        ivc.add_output('alt', val=np.zeros(self.n), units='m', desc='altitude above MSL')

        ivc.add_output('pres', val=101325.0*np.ones(self.n), units='Pa', desc='atmospheric_pressure')

        ivc.add_output('CT', val=np.linspace(0, 1.0E4, self.n), units=None,
                       desc='coefficient of thrust')

        self.p.model.add_subsystem('propulsion', PropulsionGroup(num_nodes=self.n))

        self.p.model.connect('alt', 'propulsion.alt')
        self.p.model.connect('pres', 'propulsion.pres')
        self.p.model.connect('CT', 'propulsion.CT')

        self.p.setup(force_alloc_complex=True)

        self.p.run_model()

    def test_results(self):

        assert_near_equal(self.p['propulsion.tsfc'],
                          2 * 8.951e-6 * 9.80665 * np.ones(self.n))

    def test_partials(self):
        cpd = self.p.check_partials(method='cs', out_stream=None)
        assert_check_partials(cpd)



from __future__ import print_function, division, absolute_import

import unittest

from openmdao.utils.testing_utils import use_tempdirs

try:
    import matplotlib
    matplotlib.use('Agg')
    SHOW_PLOTS = True
except ImportError:
    SHOW_PLOTS = False


@use_tempdirs
class TestLengthConstrainedBrachistochrone(unittest.TestCase):

    def test_length_constrained_brachistochrone(self):
        import openmdao.api as om
        import dymos as dm
        from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE
        from dymos.examples.length_constrained_brachistochrone.arc_length_comp import ArcLengthComp

        MAX_ARCLENGTH = 11.9
        OPTIMIZER = 'SLSQP'

        p = om.Problem(model=om.Group())
        p.add_recorder(om.SqliteRecorder('length_constrained_brach_sol.db'))

        if OPTIMIZER == 'SNOPT':
            p.driver = om.pyOptSparseDriver()
            p.driver.options['optimizer'] = OPTIMIZER
            p.driver.opt_settings['Major iterations limit'] = 1000
            p.driver.opt_settings['Major feasibility tolerance'] = 1.0E-6
            p.driver.opt_settings['Major optimality tolerance'] = 1.0E-5
            p.driver.opt_settings['iSumm'] = 6
            p.driver.opt_settings['Verify level'] = 3
        else:
            p.driver = om.ScipyOptimizeDriver()

        p.driver.declare_coloring()

        # Create the transcription so we can get the number of nodes for the downstream analysis
        tx = dm.Radau(num_segments=20, order=3, compressed=False)

        traj = dm.Trajectory()
        phase = dm.Phase(transcription=tx, ode_class=BrachistochroneODE)
        traj.add_phase('phase0', phase)

        p.model.add_subsystem('traj', traj)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', units='m', rate_source='xdot', fix_initial=True, fix_final=True)
        phase.add_state('y', units='m', rate_source='ydot', fix_initial=True, fix_final=True)
        phase.add_state('v', units='m/s', rate_source='vdot', fix_initial=True, fix_final=False)

        phase.add_control('theta', units='deg', lower=0.01, upper=179.9,
                          continuity=True, rate_continuity=True)

        phase.add_parameter('g', units='m/s**2', opt=False, val=9.80665)

        # Minimize time at the end of the phase
        phase.add_objective('time', loc='final', scaler=1)

        # p.model.options['assembled_jac_type'] = top_level_jacobian.lower()
        # p.model.linear_solver = DirectSolver(assemble_jac=True)

        # Add the arc length component
        p.model.add_subsystem('arc_length_comp',
                              subsys=ArcLengthComp(num_nodes=tx.grid_data.num_nodes))

        p.model.connect('traj.phase0.timeseries.theta', 'arc_length_comp.theta')
        p.model.connect('traj.phase0.timeseries.x', 'arc_length_comp.x')

        p.model.add_constraint('arc_length_comp.S', upper=MAX_ARCLENGTH, ref=1)

        p.setup(check=True)

        phase.set_time_val(initial=0.0, duration=2.0)
        phase.set_state_val('x', [0, 10])
        phase.set_state_val('y', [10, 5])
        phase.set_state_val('v', [0, 9.9])
        phase.set_control_val('theta', [5, 100])
        phase.set_parameter_val('g', 9.80665)

        p.run_driver()

        p.record(case_name='final')

        # Plot results
        if SHOW_PLOTS:
            import matplotlib.pyplot as plt

            # Generate the explicitly simulated trajectory
            exp_out = traj.simulate()

            # Extract the timeseries from the implicit solution and the explicit simulation
            x = p.get_val('traj.phase0.timeseries.x')
            y = p.get_val('traj.phase0.timeseries.y')
            t = p.get_val('traj.phase0.timeseries.time')
            theta = p.get_val('traj.phase0.timeseries.theta')

            x_exp = exp_out.get_val('traj.phase0.timeseries.x')
            y_exp = exp_out.get_val('traj.phase0.timeseries.y')
            t_exp = exp_out.get_val('traj.phase0.timeseries.time')
            theta_exp = exp_out.get_val('traj.phase0.timeseries.theta')

            fig, axes = plt.subplots(nrows=2, ncols=1)

            axes[0].plot(x, y, 'o')
            axes[0].plot(x_exp, y_exp, '-')
            axes[0].set_xlabel('x (m)')
            axes[0].set_ylabel('y (m)')

            axes[1].plot(t, theta, 'o')
            axes[1].plot(t_exp, theta_exp, '-')
            axes[1].set_xlabel('time (s)')
            axes[1].set_ylabel(r'$\theta$ (deg)')

            plt.show()

        return p


if __name__ == '__main__':
    unittest.main()



from openmdao.api import Group
from dymos.examples.shuttle_reentry.heating_comp import AerodynamicHeating
from dymos.examples.shuttle_reentry.flight_dynamics_comp import FlightDynamics
from dymos.examples.shuttle_reentry.aerodynamics_comp import Aerodynamics
from dymos.examples.shuttle_reentry.atmosphere_comp import Atmosphere


class ShuttleODE(Group):
    """
    The ODE for the Shuttle reentry problem.

    References
    ----------
    .. [1] Betts, John T., Practical Methods for Optimal Control and Estimation Using Nonlinear
           Programming, p. 248, 2010.
    """

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        self.add_subsystem('atmosphere', subsys=Atmosphere(num_nodes=nn),
                           promotes_inputs=['h'], promotes_outputs=['rho'])
        self.add_subsystem('aerodynamics', subsys=Aerodynamics(num_nodes=nn),
                           promotes_inputs=['alpha', 'v', 'rho'],
                           promotes_outputs=['lift', 'drag'])
        self.add_subsystem('heating', subsys=AerodynamicHeating(num_nodes=nn),
                           promotes_inputs=['rho', 'v', 'alpha'], promotes_outputs=['q'])
        self.add_subsystem('eom', subsys=FlightDynamics(num_nodes=nn),
                           promotes_inputs=['beta', 'gamma', 'h', 'psi', 'theta', 'v', 'lift',
                                            'drag'],
                           promotes_outputs=['hdot', 'gammadot', 'phidot', 'psidot', 'thetadot',
                                             'vdot'])
